from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path
from typing import Any

from utils.numpy_compat import ensure_numpy_compat

from .embeddings import embed_text

ensure_numpy_compat()

try:  # pragma: no cover - dependencies resolved at runtime
    import chromadb
except Exception as exc:  # pragma: no cover - optional dependency failure
    chromadb = None  # type: ignore[assignment]
    _IMPORT_ERROR: Exception | None = exc
else:  # pragma: no cover - chromadb available
    _IMPORT_ERROR = None


VECTOR_DB_DIR = Path(__file__).parent.parent.parent.parent / "data" / "vectors"
_CLIENT = None
_COLLECTION = None


def _get_collection():  # type: ignore[no-untyped-def]
    global _CLIENT, _COLLECTION
    if chromadb is None:
        raise RuntimeError(
            "Semantic search dependencies are not available; install the 'vector' extra"
        ) from _IMPORT_ERROR

    if _CLIENT is None:
        _CLIENT = chromadb.PersistentClient(path=str(VECTOR_DB_DIR))
    if _COLLECTION is None:
        _COLLECTION = _CLIENT.get_or_create_collection("rabbitmq_ops")
    return _COLLECTION


def semantic_search(query: str, limit: int = 10) -> list[dict[str, Any]]:
    collection = _get_collection()
    embedding = embed_text(query)
    results = collection.query(
        query_embeddings=[embedding.tolist()],
        n_results=limit,
    )
    matches: list[dict[str, Any]] = []
    ids_raw = results.get("ids") or [[]]
    distances_raw = results.get("distances") or [[]]
    metadatas_raw = results.get("metadatas") or [[]]
    documents_raw = results.get("documents") or [[]]

    ids = ids_raw[0] if ids_raw else []
    distances = distances_raw[0] if distances_raw else []
    metadatas = metadatas_raw[0] if metadatas_raw else []
    documents = documents_raw[0] if documents_raw else []

    for idx in range(len(ids)):
        matches.append(
            {
                "id": ids[idx],
                "score": distances[idx],
                "metadata": metadatas[idx] if metadatas else {},
                "document": documents[idx] if documents else None,
            }
        )
    return matches



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import pytest

from src.models.log_entry import LogCategory, LogEntry, LogLevel


def _base_payload() -> dict:
    return {
        "schema_version": "1.0.0",
        "timestamp": "2025-10-15T12:00:00Z",
        "level": LogLevel.INFO.value,
        "category": LogCategory.OPERATION.value,
        "correlation_id": "cid-123",
        "message": "operation completed",
    }


def test_old_parser_reads_logs_with_minor_version_bump() -> None:
    payload = _base_payload()
    payload["schema_version"] = "1.1.0"
    payload["context"] = {"new_optional_field": "extra info"}

    entry = LogEntry.model_validate(payload)

    assert entry.schema_version == "1.1.0"
    assert entry.context == {"new_optional_field": "extra info"}


def test_old_parser_reads_logs_with_patch_version_bump() -> None:
    payload = _base_payload()
    payload["schema_version"] = "1.0.1"

    entry = LogEntry.model_validate(payload)

    assert entry.schema_version == "1.0.1"


def test_parser_warns_on_major_version_mismatch() -> None:
    payload = _base_payload()
    payload["schema_version"] = "2.0.0"

    with pytest.raises(ValueError, match="unsupported schema_version major component"):
        LogEntry.model_validate(payload)


@pytest.mark.parametrize("invalid_version", ["1.0", "1", "v1.0.0", "1.0.0.0"])
def test_schema_version_increment_follows_semver_rules(invalid_version: str) -> None:
    payload = _base_payload()
    payload["schema_version"] = invalid_version

    with pytest.raises(ValueError):
        LogEntry.model_validate(payload)



"""Ensure the project src/ directory is importable for all Python entrypoints."""

from __future__ import annotations

import logging as _  # noqa: F401  # CRITICAL: Import stdlib logging BEFORE adding src/ to path
import sys
from pathlib import Path

SRC_PATH = str(Path(__file__).resolve().parent / "src")
if SRC_PATH not in sys.path:
    sys.path.insert(0, SRC_PATH)



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import asyncio
import contextlib
import threading
from collections import deque
from collections.abc import Awaitable, Callable, Iterable
from typing import Any

DEFAULT_QUEUE_SIZE = 10_000
DEFAULT_FLUSH_INTERVAL = 0.1
DEFAULT_BATCH_SIZE = 100


class AsyncLogWriter:
    def __init__(
        self,
        *,
        max_queue_size: int = DEFAULT_QUEUE_SIZE,
        flush_interval: float = DEFAULT_FLUSH_INTERVAL,
        batch_size: int = DEFAULT_BATCH_SIZE,
        write_batch: Callable[[Iterable[dict[str, Any]]], Awaitable[None]] | None = None,
    ) -> None:
        if max_queue_size < 1:
            raise ValueError("max_queue_size must be >= 1")
        if flush_interval <= 0:
            raise ValueError("flush_interval must be > 0")
        if batch_size < 1:
            raise ValueError("batch_size must be >= 1")

        if write_batch is None:

            async def raise_if_not_provided(_: Iterable[dict[str, Any]]) -> None:
                raise RuntimeError("write_batch callable must be provided")

            write_batch = raise_if_not_provided

        self._queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue(maxsize=max_queue_size)
        self._flush_interval = flush_interval
        self._batch_size = batch_size
        self._write_batch = write_batch

        self._worker_task: asyncio.Task[None] | None = None
        self._shutdown_event = asyncio.Event()
        self._permit = asyncio.Semaphore(max_queue_size)

        self._flush_requested = asyncio.Event()

        self._lock = asyncio.Lock()
        self._pending_entries: deque[dict[str, Any]] = deque()
        self._pending_lock = threading.Lock()

    async def start(self) -> None:
        async with self._lock:
            if self._worker_task and not self._worker_task.done():
                raise RuntimeError("AsyncLogWriter already started")

            self._shutdown_event.clear()
            self._flush_requested.clear()
            self._worker_task = asyncio.create_task(self._run())

    async def stop(self, *, timeout: float = 30.0) -> None:
        """
        Stop the async log writer and flush all pending logs.

        Args:
            timeout: Maximum time to wait for flush completion (seconds).
                     Defaults to 30 seconds to prevent indefinite hang.

        Raises:
            asyncio.TimeoutError: If flush doesn't complete within timeout.
        """
        async with self._lock:
            if not self._worker_task:
                return

            self._shutdown_event.set()
            self._flush_requested.set()

            try:
                # Wait for flush with timeout
                await asyncio.wait_for(self.flush(), timeout=timeout)
                await asyncio.wait_for(self._worker_task, timeout=max(1.0, timeout - 1.0))
            except TimeoutError:
                # Force cancel if timeout exceeded
                self._worker_task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await self._worker_task

            self._worker_task = None

    async def shutdown(self, *, timeout: float = 30.0) -> None:
        """
        Gracefully shutdown the writer, flushing all buffered logs.

        This is an alias for stop() for clearer semantics.

        Args:
            timeout: Maximum time to wait for flush completion (seconds).
        """
        await self.stop(timeout=timeout)

    async def write_log(self, entry: dict[str, Any]) -> None:
        # Track pending entries for synchronous flush fallbacks
        with self._pending_lock:
            self._pending_entries.append(dict(entry))
        await self._permit.acquire()
        await self._queue.put(entry)

    async def flush(self) -> None:
        self._flush_requested.set()
        await self._queue.join()

    async def _run(self) -> None:
        batch: list[dict[str, Any]] = []

        get_task: asyncio.Task[dict[str, Any]] | None = None
        sleep_task: asyncio.Task[Any] | None = None
        flush_task: asyncio.Task[Any] | None = None

        def _ensure_get_task() -> None:
            nonlocal get_task
            if get_task is None or get_task.done():
                get_task = asyncio.create_task(self._queue.get())

        def _restart_sleep_task() -> None:
            nonlocal sleep_task
            if sleep_task is not None:
                sleep_task.cancel()
            sleep_task = asyncio.create_task(asyncio.sleep(self._flush_interval))

        def _ensure_flush_task() -> None:
            nonlocal flush_task
            if flush_task is None or flush_task.done():
                flush_task = asyncio.create_task(self._flush_requested.wait())

        _ensure_get_task()
        _restart_sleep_task()
        _ensure_flush_task()

        flush_pending = False

        try:
            while True:
                if self._shutdown_event.is_set() and self._queue.empty() and not batch:
                    break

                wait_tasks = {t for t in (get_task, sleep_task, flush_task) if t is not None}
                if not wait_tasks:
                    await asyncio.sleep(self._flush_interval)
                    continue

                done, _ = await asyncio.wait(wait_tasks, return_when=asyncio.FIRST_COMPLETED)

                if get_task in done and get_task is not None:
                    item = get_task.result()
                    batch.append(item)
                    if len(batch) == 1:
                        _restart_sleep_task()
                    if len(batch) >= self._batch_size:
                        _restart_sleep_task()
                        await self._emit_batch(batch)
                        if flush_pending and self._queue.empty():
                            flush_pending = False
                    if flush_pending and self._queue.empty():
                        if batch:
                            _restart_sleep_task()
                            await self._emit_batch(batch)
                        flush_pending = False
                    _ensure_get_task()

                if sleep_task in done and sleep_task is not None:
                    _restart_sleep_task()
                    if batch:
                        await self._emit_batch(batch)
                        if flush_pending and self._queue.empty():
                            flush_pending = False

                if flush_task in done and flush_task is not None:
                    self._flush_requested.clear()
                    _ensure_flush_task()
                    flush_pending = True
                    if self._queue.empty():
                        if batch:
                            _restart_sleep_task()
                            await self._emit_batch(batch)
                        flush_pending = False
        finally:
            for task in (get_task, sleep_task, flush_task):
                if task is None:
                    continue
                task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await task

            if batch:
                await self._emit_batch(batch)

    async def _emit_batch(self, batch: list[dict[str, Any]]) -> None:
        if not batch:
            return

        to_write = list(batch)
        try:
            await self._write_batch(to_write)
            if to_write:
                with self._pending_lock:
                    for _ in to_write:
                        if self._pending_entries:
                            self._pending_entries.popleft()
        finally:
            for _ in to_write:
                self._queue.task_done()
                self._permit.release()
        batch.clear()

    def drain_pending(self) -> list[dict[str, Any]]:
        """Return any entries that were queued but not written yet."""
        with self._pending_lock:
            if not self._pending_entries:
                return []
            pending = list(self._pending_entries)
            self._pending_entries.clear()
            return pending



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from collections.abc import Iterable
from copy import deepcopy

import pytest

from src.logging.redaction import apply_redaction, apply_redaction_to_event


@pytest.fixture
def base_event():
    return {
        "message": "Connecting to broker",
        "connection": "amqp://user:supersecret@localhost",
        "password": "plainpassword",
        "api_key": "abc123",
        "token": "xyz987",
        "authorization": "Bearer token-value",
    }


def test_apply_redaction_does_not_mutate_original(base_event):
    original = deepcopy(base_event)
    apply_redaction(base_event)
    assert base_event == original


def test_redact_password_in_connection_string(base_event):
    redacted = apply_redaction(base_event)
    assert redacted["connection"] == "amqp://user:[REDACTED]@localhost"


def test_redact_password_field(base_event):
    redacted = apply_redaction(base_event)
    assert redacted["password"] == "[REDACTED]"


def test_redact_api_key(base_event):
    redacted = apply_redaction(base_event)
    assert redacted["api_key"] == "[REDACTED]"


def test_redact_token(base_event):
    redacted = apply_redaction(base_event)
    assert redacted["token"] == "[REDACTED]"


def test_redact_bearer_token(base_event):
    redacted = apply_redaction(base_event)
    assert redacted["authorization"] == "Bearer [REDACTED]"


def test_redact_amqp_password_preserves_username():
    event = {"connection": "amqp://important-user:secret-pass@localhost"}
    redacted = apply_redaction(event)
    assert redacted["connection"].startswith("amqp://important-user:")
    assert redacted["connection"] == "amqp://important-user:[REDACTED]@localhost"


def test_redaction_works_in_nested_context():
    event = {
        "outer": {
            "token": "nested-secret",
            "details": ["no-secret", "Bearer abc"],
            "connection": "amqp://user:nestedpass@host",
        }
    }

    redacted = apply_redaction(event)

    assert redacted["outer"]["token"] == "[REDACTED]"
    assert redacted["outer"]["details"][1] == "Bearer [REDACTED]"
    assert redacted["outer"]["connection"] == "amqp://user:[REDACTED]@host"


def test_redact_keys_containing_token_suffix():
    event = {
        "secondary_token": "should-hide",
        "refresh_token": "also-hidden",
    }

    redacted = apply_redaction_to_event(event)

    assert redacted["secondary_token"] == "[REDACTED]"
    assert redacted["refresh_token"] == "[REDACTED]"


def test_no_credentials_in_log_output(base_event):
    redacted = apply_redaction(base_event)

    sensitive_patterns: Iterable[str] = ["plainpassword", "abc123", "xyz987", "token-value"]
    serialized = " ".join(str(value) for value in redacted.values())
    for pattern in sensitive_patterns:
        assert pattern not in serialized


def test_apply_redaction_keeps_unrelated_fields():
    event = {"message": "Everything is fine", "count": 42}
    redacted = apply_redaction(event)
    assert redacted == event



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""CLI entry point for rabbitmq-mcp-server."""

import click

from cli.commands.binding import binding
from cli.commands.exchange import exchange
from cli.commands.queue import queue


@click.group()
def cli() -> None:
    """Interact with RabbitMQ topology operations."""


cli.add_command(queue)
cli.add_command(exchange)
cli.add_command(binding)


__all__ = ["cli"]



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

from .parser import OpenAPIParseError, parse_openapi_spec


class OperationMetadata:
    def __init__(self, operation_id: str, path: str, method: str, spec: dict[str, Any]):
        self.operation_id = operation_id
        self.path = path
        self.method = method
        self.spec = spec


class OperationRegistry:
    def __init__(self, openapi_paths: list[str]):
        self._operations: dict[str, OperationMetadata] = {}
        self._loaded = False
        self._openapi_paths = openapi_paths
        self._load_operations()

    def _load_operations(self) -> None:
        for path in self._openapi_paths:
            try:
                spec = parse_openapi_spec(path)
            except OpenAPIParseError as e:
                raise RuntimeError(f"Failed to load OpenAPI spec {path}: {e}")
            for api_path, methods in spec.get("paths", {}).items():
                for method, op in methods.items():
                    op_id = op.get("operationId")
                    if op_id:
                        self._operations[op_id] = OperationMetadata(
                            operation_id=op_id, path=api_path, method=method, spec=op
                        )
        self._loaded = True

    def get_operation(self, operation_id: str) -> OperationMetadata | None:
        return self._operations.get(operation_id)

    def list_operations(self, tag: str | None = None) -> list[OperationMetadata]:
        if tag:
            return [op for op in self._operations.values() if tag in op.spec.get("tags", [])]
        return list(self._operations.values())



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from cli.commands import binding as binding_cmd
from cli.commands import exchange as exchange_cmd
from cli.commands import queue as queue_cmd
from cli.commands.shared import format_table_output


def _header_sequence(output: str) -> list[str]:
    header_line = output.splitlines()[0]
    return [part.strip() for part in header_line.strip().strip("|").split("|")]


def test_queue_table_column_ordering() -> None:
    payload: list[dict[str, object]] = [
        {
            "name": "orders.queue",
            "vhost": "/",
            "messages": 10,
            "consumers": 3,
            "state": "running",
            "messages_ready": 8,
            "messages_unacknowledged": 2,
            "memory": 1024,
        }
    ]

    base_output = format_table_output(payload, queue_cmd.BASE_COLUMNS, footer="footer")
    verbose_output = format_table_output(payload, queue_cmd.VERBOSE_COLUMNS, footer="footer")

    assert _header_sequence(base_output) == [header for header, _ in queue_cmd.BASE_COLUMNS]
    assert _header_sequence(verbose_output) == [header for header, _ in queue_cmd.VERBOSE_COLUMNS]


def test_exchange_table_column_ordering() -> None:
    payload: list[dict[str, object]] = [
        {
            "name": "orders.exchange",
            "vhost": "/",
            "type": "topic",
            "durable": True,
            "auto_delete": False,
            "internal": False,
            "bindings_count": 2,
            "message_stats": {"publish_in": 5, "publish_out": 5},
        }
    ]

    base_output = format_table_output(payload, exchange_cmd.BASE_COLUMNS, footer="footer")
    verbose_output = format_table_output(payload, exchange_cmd.VERBOSE_COLUMNS, footer="footer")

    assert _header_sequence(base_output) == [header for header, _ in exchange_cmd.BASE_COLUMNS]
    assert _header_sequence(verbose_output) == [
        header for header, _ in exchange_cmd.VERBOSE_COLUMNS
    ]


def test_binding_table_column_ordering() -> None:
    payload: list[dict[str, object]] = [
        {
            "source": "orders.exchange",
            "destination": "orders.queue",
            "destination_type": "queue",
            "routing_key": "order.created",
            "vhost": "/",
        }
    ]

    output = format_table_output(payload, binding_cmd.BASE_COLUMNS, footer="footer")

    assert _header_sequence(output) == [header for header, _ in binding_cmd.BASE_COLUMNS]



"""
This file is auto-generated from OpenAPI contracts.
LGPL v3.0 applies. See LICENSE in project root.
"""

from pydantic import BaseModel, Field


class PaginationMetadata(BaseModel):
    page: int = Field(...)
    pageSize: int = Field(...)
    totalItems: int = Field(...)
    totalPages: int = Field(...)
    hasNextPage: bool = Field(...)
    hasPreviousPage: bool = Field(...)



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

import pytest

from utils import validation
from utils.errors import AuthorizationError, ConnectionError, NotFoundError, ValidationError


class FakeExecutor:
    def __init__(self) -> None:
        self.calls: dict[str, int] = {}
        self.base_url = "http://localhost:15672/api"
        self.vhost = "/"
        self.raise_error: Exception | None = None

    def _encode_vhost(self, vhost: str) -> str:
        return vhost.replace("/", "%2F")

    def _request(self, method: str, endpoint: str, **_: Any) -> Any:
        key = f"{method} {endpoint}"
        self.calls[key] = self.calls.get(key, 0) + 1
        if self.raise_error:
            raise self.raise_error
        return {}


def setup_function(_: Any) -> None:
    validation.reset_vhost_cache()


def test_validate_name_success() -> None:
    validation.validate_name("orders.queue-1")


@pytest.mark.parametrize("value", ["orders queue", "", "!invalid", "a" * 256])
def test_validate_name_failure(value: str) -> None:
    with pytest.raises(ValidationError) as exc:
        validation.validate_name(value)
    assert exc.value.code == "INVALID_NAME"
    assert exc.value.field == "name"


@pytest.mark.parametrize(
    "exchange_type",
    ["direct", "topic", "fanout", "headers"],
)
def test_validate_exchange_type_success(exchange_type: str) -> None:
    validation.validate_exchange_type(exchange_type)


def test_validate_exchange_type_failure() -> None:
    with pytest.raises(ValidationError) as exc:
        validation.validate_exchange_type("invalid")
    assert exc.value.code == "INVALID_EXCHANGE_TYPE"
    assert exc.value.field == "type"


@pytest.mark.parametrize(
    "exchange_type, routing_key",
    [
        ("direct", "orders.*"),
        ("fanout", "#"),
        ("headers", "*"),
    ],
)
def test_validate_routing_key_invalid_for_non_topic(exchange_type: str, routing_key: str) -> None:
    with pytest.raises(ValidationError) as exc:
        validation.validate_routing_key(routing_key, exchange_type)
    assert exc.value.code == "INVALID_ROUTING_KEY"


@pytest.mark.parametrize(
    "exchange_type, routing_key",
    [
        ("topic", "orders.*"),
        ("topic", "orders.#"),
        ("direct", "orders.created"),
    ],
)
def test_validate_routing_key_allowed(exchange_type: str, routing_key: str) -> None:
    validation.validate_routing_key(routing_key, exchange_type)


def test_validate_vhost_exists_uses_cache() -> None:
    executor = FakeExecutor()
    validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]
    validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]

    # Only one HTTP call should happen due to 60s cache
    assert executor.calls["GET /vhosts/%2F"] == 1


def test_validate_vhost_exists_not_found() -> None:
    executor = FakeExecutor()
    executor.raise_error = NotFoundError(
        code="NOT_FOUND",
        message="Virtual host missing",
        action="Create vhost",
        context={"status": 404},
    )

    with pytest.raises(ValidationError) as exc:
        validation.validate_vhost_exists("/missing", executor)  # type: ignore[arg-type]

    assert exc.value.code == "VHOST_NOT_FOUND"


def test_validate_vhost_exists_auth_error_invalidates_cache() -> None:
    executor = FakeExecutor()
    auth_error = AuthorizationError(
        code="UNAUTHORIZED",
        message="Access denied",
        action="Update credentials",
    )
    executor.raise_error = auth_error

    with pytest.raises(AuthorizationError):
        validation.validate_vhost_exists("/secure", executor)  # type: ignore[arg-type]

    executor.raise_error = None
    validation.validate_vhost_exists("/secure", executor)  # type: ignore[arg-type]
    assert executor.calls["GET /vhosts/%2Fsecure"] == 2


def test_reset_vhost_cache_clears_cached_results() -> None:
    executor = FakeExecutor()
    validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]
    validation.reset_vhost_cache()

    executor.raise_error = NotFoundError(
        code="NOT_FOUND",
        message="missing",
        action="create",
    )

    with pytest.raises(ValidationError) as exc:
        validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]

    assert exc.value.code == "VHOST_NOT_FOUND"


def test_validate_vhost_exists_propagates_connection_error() -> None:
    executor = FakeExecutor()
    connection_error = ConnectionError(
        code="NETWORK_ERROR",
        message="offline",
        action="Retry",
    )
    executor.raise_error = connection_error

    with pytest.raises(ConnectionError) as exc:
        validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]

    assert exc.value is connection_error


def test_validate_vhost_exists_wraps_unexpected_exception() -> None:
    executor = FakeExecutor()
    executor.raise_error = RuntimeError("boom")

    with pytest.raises(ConnectionError) as exc:
        validation.validate_vhost_exists("/", executor)  # type: ignore[arg-type]

    assert exc.value.code == "CONNECTION_ERROR"



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Implementação do MCP tool search-ids."""

from pathlib import Path

from rabbitmq_mcp_connection.schemas.mcp import (
    MCPError,
    MCPToolResult,
    PaginationMetadata,
    PaginationParams,
    SearchItem,
    SearchResult,
)
from rabbitmq_mcp_connection.tools.contracts import load_connection_operations
from utils.numpy_compat import ensure_numpy_compat

ensure_numpy_compat()

try:  # pragma: no cover - dependências opcionais
    import chromadb
    from sentence_transformers import SentenceTransformer
except Exception as exc:  # pragma: no cover - fallback sem embeddings
    chromadb = None  # type: ignore[assignment]
    SentenceTransformer = None
    _IMPORT_ERROR: Exception | None = exc
else:  # pragma: no cover - dependências disponíveis
    _IMPORT_ERROR = None

DATA_DIR = Path(__file__).resolve().parents[4] / "data" / "vectors"
COLLECTION_NAME = "connection_operations"


class SearchBackend:
    def search(self, query: str, limit: int) -> list[tuple[str, float, dict]]:
        raise NotImplementedError


class KeywordFallbackBackend(SearchBackend):
    def __init__(self) -> None:
        self.operations = load_connection_operations()

    def search(self, query: str, limit: int) -> list[tuple[str, float, dict]]:
        query_lower = query.lower()
        scored: list[tuple[str, float, dict]] = []
        for operation_id, schema in self.operations.items():
            text = f"{schema.name} {schema.description}".lower()
            score = 1.0 if query_lower and query_lower in text else 0.25
            metadata = {
                "name": schema.name,
                "category": schema.category,
                "description": schema.description,
            }
            scored.append((operation_id, score, metadata))
        scored.sort(key=lambda item: item[1], reverse=True)
        return scored[:limit]


class ChromaSearchBackend(SearchBackend):
    def __init__(self) -> None:
        if chromadb is None or SentenceTransformer is None:
            raise RuntimeError("Dependências de busca semântica não instaladas")

        self._base_path = DATA_DIR
        self._base_path.mkdir(parents=True, exist_ok=True)
        self._client = chromadb.PersistentClient(path=str(self._base_path))
        self._collection = self._client.get_or_create_collection(COLLECTION_NAME)
        self._model = SentenceTransformer("all-MiniLM-L6-v2")

    def search(self, query: str, limit: int) -> list[tuple[str, float, dict]]:
        embedding = self._model.encode(query)
        result = self._collection.query(
            query_embeddings=[embedding.tolist()],
            n_results=limit,
        )

        ids_raw = result.get("ids") or [[]]
        distances_raw = result.get("distances") or [[]]
        metadatas_raw = result.get("metadatas") or [[]]

        ids = ids_raw[0] if ids_raw else []
        distances = distances_raw[0] if distances_raw else []
        metadatas = metadatas_raw[0] if metadatas_raw else []

        scored: list[tuple[str, float, dict]] = []
        for op_id, distance, metadata in zip(ids, distances, metadatas):
            score = float(max(0.0, 1.0 - distance))
            scored.append((op_id, score, dict(metadata or {})))
        return scored


def _resolve_backend() -> SearchBackend:
    if chromadb is not None and SentenceTransformer is not None:
        try:
            return ChromaSearchBackend()
        except Exception:  # pragma: no cover - fallback
            pass
    return KeywordFallbackBackend()


_BACKEND = _resolve_backend()


def handle_search_ids(query: str, pagination: PaginationParams | None = None) -> MCPToolResult:
    if not query:
        return MCPToolResult(
            success=False,
            error=MCPError(code="INVALID_QUERY", message="Query must be non-empty"),
        )

    pagination = pagination or PaginationParams()
    backend = _BACKEND
    raw_results = backend.search(query, pagination.page_size * pagination.page)

    start_index = (pagination.page - 1) * pagination.page_size
    page_items = raw_results[start_index : start_index + pagination.page_size]

    items = [
        SearchItem(operation_id=op_id, score=score, metadata=metadata)
        for op_id, score, metadata in page_items
    ]

    result = SearchResult(
        items=items,
        pagination=PaginationMetadata(
            page=pagination.page,
            page_size=pagination.page_size,
            total=len(raw_results),
        ),
    )

    metadata = {"backend": backend.__class__.__name__}
    if isinstance(backend, KeywordFallbackBackend) and _IMPORT_ERROR is not None:
        metadata["dependency_error"] = str(_IMPORT_ERROR)

    return MCPToolResult(
        success=True,
        result=result.model_dump(),
        metadata=metadata,
    )


__all__ = ["handle_search_ids"]



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""MCP call-id tool dispatcher for topology list operations."""

from collections.abc import Callable
from typing import Any

from pydantic import BaseModel, Field, ValidationError

from config.settings import Settings
from tools.operations.bindings import (
    Binding,
    PaginatedBindingResponse,
    create_binding,
    delete_binding,
    list_bindings,
)
from tools.operations.exchanges import (
    Exchange,
    ExchangeOptions,
    PaginatedExchangeResponse,
    create_exchange,
    delete_exchange,
    list_exchanges,
)
from tools.operations.queues import (
    PaginatedQueueResponse,
    Queue,
    QueueOptions,
    create_queue,
    delete_queue,
    list_queues,
)
from utils.errors import RabbitMQError, format_error


class PaginationInput(BaseModel):
    page: int = Field(default=1, ge=1)
    pageSize: int = Field(default=50, ge=1, le=200)


class ConnectionParams(BaseModel):
    host: str
    user: str
    password: str
    port: int = 15672
    use_tls: bool = False
    connection_vhost: str = Field(default="/", alias="connection_vhost")
    vhost: str | None = None

    model_config = {"populate_by_name": True}


class QueueCreateInput(ConnectionParams):
    name: str
    durable: bool = False
    exclusive: bool = False
    auto_delete: bool = False
    arguments: dict[str, Any] = Field(default_factory=dict)

    model_config = {"populate_by_name": True, "extra": "forbid"}


class ExchangeCreateInput(ConnectionParams):
    name: str
    type: str
    durable: bool = False
    auto_delete: bool = False
    internal: bool = False
    arguments: dict[str, Any] = Field(default_factory=dict)

    model_config = {"populate_by_name": True, "extra": "forbid"}


class BindingCreateInput(ConnectionParams):
    exchange: str
    queue: str
    routing_key: str = ""
    arguments: dict[str, Any] = Field(default_factory=dict)

    model_config = {"populate_by_name": True, "extra": "forbid"}


class QueueDeleteInput(ConnectionParams):
    name: str
    force: bool = False

    model_config = {"populate_by_name": True, "extra": "forbid"}


class ExchangeDeleteInput(ConnectionParams):
    name: str

    model_config = {"populate_by_name": True, "extra": "forbid"}


class BindingDeleteInput(ConnectionParams):
    exchange: str
    queue: str
    properties_key: str

    model_config = {"populate_by_name": True, "extra": "forbid"}


ResponseModel = (
    PaginatedQueueResponse
    | PaginatedExchangeResponse
    | PaginatedBindingResponse
    | Queue
    | Exchange
    | Binding
    | dict[str, Any]
)


_LIST_HANDLERS: dict[str, Callable[..., ResponseModel]] = {
    "queues.list": list_queues,
    "exchanges.list": list_exchanges,
    "bindings.list": list_bindings,
}


def _build_settings(params: ConnectionParams) -> Settings:
    return Settings(
        host=params.host,
        port=params.port,
        user=params.user,
        password=params.password,
        vhost=params.connection_vhost,
        use_tls=params.use_tls,
    )


def call_id(
    endpoint_id: str,
    params: dict[str, Any] | None = None,
    pagination: dict[str, Any] | None = None,
) -> ResponseModel:
    handler = _LIST_HANDLERS.get(endpoint_id)
    if handler is not None:
        try:
            connection_params = ConnectionParams.model_validate(params or {})
            pagination_params = PaginationInput.model_validate(pagination or {})
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(connection_params)

        try:
            return handler(
                vhost=connection_params.vhost,
                page=pagination_params.page,
                pageSize=pagination_params.pageSize,
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "queues.create":
        try:
            payload = QueueCreateInput.model_validate(params or {})
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        options = QueueOptions(
            durable=payload.durable,
            exclusive=payload.exclusive,
            auto_delete=payload.auto_delete,
            arguments=payload.arguments,
        )
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return create_queue(
                vhost=target_vhost,
                name=payload.name,
                options=options,
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "exchanges.create":
        try:
            payload = ExchangeCreateInput.model_validate(params or {})  # type: ignore[assignment]
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        options = ExchangeOptions(  # type: ignore[assignment]
            durable=payload.durable,  # type: ignore[attr-defined]
            auto_delete=payload.auto_delete,
            internal=payload.internal,  # type: ignore[attr-defined]
            arguments=payload.arguments,
        )
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return create_exchange(
                vhost=target_vhost,
                name=payload.name,
                exchange_type=payload.type,  # type: ignore[attr-defined]
                options=options,  # type: ignore[arg-type]
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "bindings.create":
        try:
            payload = BindingCreateInput.model_validate(params or {})  # type: ignore[assignment]
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return create_binding(
                vhost=target_vhost,
                exchange=payload.exchange,  # type: ignore[attr-defined]
                queue=payload.queue,  # type: ignore[attr-defined]
                routing_key=payload.routing_key,  # type: ignore[attr-defined]
                args=payload.arguments,
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "queues.delete":
        try:
            payload = QueueDeleteInput.model_validate(params or {})  # type: ignore[assignment]
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return delete_queue(
                vhost=target_vhost,
                name=payload.name,
                force=payload.force,  # type: ignore[attr-defined]
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "exchanges.delete":
        try:
            payload = ExchangeDeleteInput.model_validate(params or {})  # type: ignore[assignment]
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return delete_exchange(
                vhost=target_vhost,
                name=payload.name,
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    if endpoint_id == "bindings.delete":
        try:
            payload = BindingDeleteInput.model_validate(params or {})  # type: ignore[assignment]
        except ValidationError as exc:
            raise format_error(exc) from exc

        settings = _build_settings(payload)
        target_vhost = payload.vhost or payload.connection_vhost

        try:
            return delete_binding(
                vhost=target_vhost,
                exchange=payload.exchange,  # type: ignore[attr-defined]
                queue=payload.queue,  # type: ignore[attr-defined]
                properties_key=payload.properties_key,  # type: ignore[attr-defined]
                settings=settings,
            )
        except RabbitMQError:
            raise
        except Exception as exc:  # pragma: no cover - defensive safety
            raise format_error(exc) from exc

    raise ValueError(f"Unsupported endpoint '{endpoint_id}'")


__all__ = ["call_id", "PaginationInput", "ConnectionParams"]



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import contextvars
import secrets
import time
import uuid

__all__ = [
    "generate_correlation_id",
    "set_correlation_id",
    "get_correlation_id",
    "reset_correlation_id",
]


_correlation_var: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "correlation_id", default=None
)


def generate_correlation_id() -> str:
    try:
        return str(uuid.uuid4())
    except Exception:
        timestamp = int(time.time() * 1_000_000)
        random_part = secrets.token_hex(3)
        return f"fallback-{timestamp}-{random_part}"


def set_correlation_id(value: str | None = None) -> str:
    correlation_id = value or generate_correlation_id()
    _correlation_var.set(correlation_id)
    return correlation_id


def get_correlation_id() -> str | None:
    return _correlation_var.get()


def reset_correlation_id() -> None:
    _correlation_var.set(None)



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Despachante para o MCP call-id tool."""

import time
from collections.abc import Awaitable, Callable
from typing import Any

from pydantic import ValidationError

from rabbitmq_mcp_connection.connection.config import load_config
from rabbitmq_mcp_connection.connection.exceptions import (
    AuthenticationError,
    ConnectionError,
    ConnectionTimeoutError,
    VHostNotFoundError,
)
from rabbitmq_mcp_connection.connection.health import HealthChecker
from rabbitmq_mcp_connection.connection.manager import ConnectionManager
from rabbitmq_mcp_connection.connection.pool import ConnectionPool
from rabbitmq_mcp_connection.logging.config import get_logger
from rabbitmq_mcp_connection.schemas.connection import (
    ConnectionConfig,
    ConnectResult,
    DisconnectResult,
    HealthStatus,
    PoolStats,
)
from rabbitmq_mcp_connection.schemas.mcp import MCPError, MCPToolResult
from rabbitmq_mcp_connection.tools.contracts import load_connection_operations

LOGGER = get_logger(__name__)


class ToolState:
    def __init__(self) -> None:
        self._config: ConnectionConfig | None = None
        self._manager: ConnectionManager | None = None
        self._health_checker: HealthChecker | None = None
        self._pool: ConnectionPool | None = None

    async def get_manager(self, overrides: dict[str, Any] | None = None) -> ConnectionManager:
        config = self._resolve_config(overrides or {})
        if self._manager is None:
            self._manager = ConnectionManager(config)
            self._config = config
            self._health_checker = HealthChecker(self._manager)
            return self._manager

        if self._config != config:
            await self._manager.disconnect()
            self._manager = ConnectionManager(config)
            self._config = config
            self._health_checker = HealthChecker(self._manager)
        return self._manager

    def get_health_checker(self, manager: ConnectionManager) -> HealthChecker:
        if self._health_checker is None or self._manager is not manager:
            self._health_checker = HealthChecker(manager)
        return self._health_checker

    def get_pool(self, config: ConnectionConfig) -> ConnectionPool:
        if self._pool is None or self._config != config:
            self._pool = ConnectionPool(config)
        return self._pool

    def peek_manager(self) -> ConnectionManager | None:
        return self._manager

    def _resolve_config(self, overrides: dict[str, Any]) -> ConnectionConfig:
        try:
            config = load_config(**overrides)
        except ValidationError as exc:
            raise ValueError(exc.errors()) from exc
        return config


_STATE = ToolState()
_OPERATIONS = load_connection_operations()


def _error(code: str, message: str, details: dict[str, Any] | None = None) -> MCPToolResult:
    return MCPToolResult(success=False, error=MCPError(code=code, message=message, details=details))


async def handle_call_id(operation_id: str, payload: dict[str, Any] | None = None) -> MCPToolResult:
    if operation_id not in _OPERATIONS:
        return _error("UNKNOWN_OPERATION", f"Operação não suportada: {operation_id}")

    payload = payload or {}

    handlers: dict[str, Callable[[dict[str, Any]], Awaitable[MCPToolResult]]] = {
        "connection.connect": _handle_connect,
        "connection.disconnect": _handle_disconnect,
        "connection.health_check": _handle_health_check,
        "connection.get_status": _handle_get_status,
        "pool.get_stats": _handle_pool_stats,
    }

    handler = handlers.get(operation_id)
    if handler is None:
        return _error("NOT_IMPLEMENTED", f"Handler não implementado para {operation_id}")

    try:
        return await handler(payload)
    except ValueError as exc:
        return _error("INVALID_PAYLOAD", "Dados inválidos", details={"errors": str(exc)})
    except ConnectionTimeoutError as exc:
        return _error("CONNECTION_TIMEOUT", str(exc))
    except AuthenticationError as exc:
        return _error("AUTHENTICATION_FAILED", str(exc))
    except VHostNotFoundError as exc:
        return _error("VHOST_NOT_FOUND", str(exc))
    except ConnectionError as exc:
        return _error("CONNECTION_ERROR", str(exc))
    except Exception as exc:  # pragma: no cover - proteção geral
        LOGGER.exception("call_id.unhandled_exception", operation=operation_id)
        return _error("INTERNAL_ERROR", "Unexpected error", details={"error": str(exc)})


async def _handle_connect(payload: dict[str, Any]) -> MCPToolResult:
    manager = await _STATE.get_manager(payload)
    start = time.perf_counter()
    await manager.connect()
    latency_ms = (time.perf_counter() - start) * 1000

    result = ConnectResult(
        connected=manager.state.is_connected,
        connection_url=manager.config.get_connection_url(),
        latency_ms=latency_ms,
        server_properties=getattr(manager.connection, "server_properties", None),
    )
    return MCPToolResult(success=True, result=result.model_dump())


async def _handle_disconnect(payload: dict[str, Any]) -> MCPToolResult:
    manager = _STATE.peek_manager()
    if manager is None or not manager.state.is_connected:
        return _error("NOT_CONNECTED", "Nenhuma conexão ativa para desconectar")

    force = bool(payload.get("force", False))
    if force:
        if manager.connection and not manager.connection.is_closed:
            await manager.connection.close()
    start = time.perf_counter()
    await manager.disconnect()
    duration_ms = (time.perf_counter() - start) * 1000
    result = DisconnectResult(disconnected=True, graceful=not force, duration_ms=duration_ms)
    return MCPToolResult(success=True, result=result.model_dump())


async def _handle_health_check(payload: dict[str, Any]) -> MCPToolResult:
    manager = await _STATE.get_manager({})
    checker = _STATE.get_health_checker(manager)
    status: HealthStatus = await checker.check_health()
    return MCPToolResult(success=True, result=status.model_dump())


async def _handle_get_status(payload: dict[str, Any]) -> MCPToolResult:
    manager = await _STATE.get_manager({})
    status = await manager.get_status()
    return MCPToolResult(success=True, result=status.model_dump())


async def _handle_pool_stats(payload: dict[str, Any]) -> MCPToolResult:
    manager = await _STATE.get_manager({})
    pool = _STATE.get_pool(manager.config)
    stats: PoolStats = await pool.get_stats()
    return MCPToolResult(success=True, result=stats.model_dump())


def reset_tool_state() -> None:
    """Reinicia o estado global (útil em testes)."""

    global _STATE
    _STATE = ToolState()


__all__ = ["handle_call_id", "reset_tool_state"]



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import asyncio
from unittest import mock

import pytest

from src.utils.async_writer import AsyncLogWriter


@pytest.mark.asyncio
async def test_async_log_writer_writes_items_in_order():
    sink = []

    async def fake_write(batch):
        sink.extend(batch)

    writer = AsyncLogWriter(
        max_queue_size=2, batch_size=1, flush_interval=0.1, write_batch=fake_write
    )
    await writer.start()

    await writer.write_log({"id": 1})
    await writer.write_log({"id": 2})

    await asyncio.wait_for(writer.flush(), timeout=1)
    await writer.stop()

    assert sink == [{"id": 1}, {"id": 2}]


@pytest.mark.asyncio
async def test_async_log_writer_blocks_when_queue_full(monkeypatch):
    sink = []
    block_event = asyncio.Event()

    async def slow_write(batch):
        sink.extend(batch)
        await block_event.wait()

    writer = AsyncLogWriter(
        max_queue_size=1, batch_size=1, flush_interval=0.1, write_batch=slow_write
    )

    original_get = writer._queue.get

    async def gated_get():  # type: ignore[override]
        await block_event.wait()
        return await original_get()

    monkeypatch.setattr(writer._queue, "get", gated_get)

    await writer.start()

    first_task = asyncio.create_task(writer.write_log({"id": "first"}))
    second_task = asyncio.create_task(writer.write_log({"id": "second"}))

    await asyncio.sleep(0)
    assert first_task.done()  # primeiro item pode entrar na fila
    assert not second_task.done()  # segundo deve aguardar capacidade

    block_event.set()
    await asyncio.wait_for(first_task, timeout=1)
    await asyncio.wait_for(second_task, timeout=1)
    await asyncio.wait_for(writer.flush(), timeout=1)
    await writer.stop()

    assert sink == [{"id": "first"}, {"id": "second"}]


@pytest.mark.asyncio
async def test_async_log_writer_batches_entries():
    batches = []

    async def capture_write(batch):
        batches.append(list(batch))

    writer = AsyncLogWriter(
        max_queue_size=10, batch_size=3, flush_interval=0.1, write_batch=capture_write
    )
    await writer.start()

    await writer.write_log({"id": 1})
    await writer.write_log({"id": 2})
    await writer.write_log({"id": 3})

    await asyncio.wait_for(writer.flush(), timeout=1)
    await writer.stop()

    assert batches == [[{"id": 1}, {"id": 2}, {"id": 3}]]


@pytest.mark.asyncio
async def test_async_log_writer_flush_interval_triggers():
    batches = []

    async def capture(batch):
        batches.extend(batch)

    writer = AsyncLogWriter(
        max_queue_size=10, batch_size=10, flush_interval=0.05, write_batch=capture
    )
    await writer.start()

    await writer.write_log({"id": 1})
    await asyncio.sleep(0.1)  # allow flush interval to trigger

    await asyncio.wait_for(writer.stop(), timeout=1)

    assert batches == [{"id": 1}]


@pytest.mark.asyncio
async def test_async_log_writer_stop_drains_queue():
    sink = []

    async def capture(batch):
        sink.extend(batch)

    writer = AsyncLogWriter(
        max_queue_size=10, batch_size=5, flush_interval=0.1, write_batch=capture
    )
    await writer.start()

    for i in range(7):
        await writer.write_log({"id": i})

    await asyncio.wait_for(writer.stop(), timeout=1)

    assert len(sink) == 7


@pytest.mark.asyncio
async def test_async_log_writer_prevents_multiple_starts():
    writer = AsyncLogWriter(write_batch=mock.AsyncMock())
    await writer.start()

    with pytest.raises(RuntimeError):
        await writer.start()

    await writer.stop()


@pytest.mark.asyncio
async def test_async_log_writer_handles_stop_without_start():
    writer = AsyncLogWriter(write_batch=mock.AsyncMock())

    # stop before start should be harmless
    await writer.stop()



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import os
import uuid
from logging.handlers import TimedRotatingFileHandler  # type: ignore[attr-defined]
from typing import Any

import structlog

LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO").upper()
LOG_FILE = os.getenv("LOG_FILE", "./logs/rabbitmq-mcp.log")

# Sensitive data patterns to redact
SENSITIVE_PATTERNS = [r"password", r"token", r"credential", r"Authorization"]


class RedactProcessor:
    def __call__(self, logger: Any, method_name: str, event_dict: dict[str, Any]) -> dict[str, Any]:
        for key in event_dict:
            if any(p in key.lower() for p in SENSITIVE_PATTERNS):
                event_dict[key] = "REDACTED"
        return event_dict


def get_logger(name: str) -> Any:
    return structlog.get_logger(name)


# Configure structlog
structlog.configure(
    processors=[
        structlog.processors.TimeStamper(fmt="iso", utc=True),
        structlog.processors.add_log_level,
        structlog.processors.UUIDAdder("correlation_id", uuid.uuid4),  # type: ignore[attr-defined]
        RedactProcessor(),
        structlog.processors.JSONRenderer(),
    ],
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

# File-based logging with rotation
file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight", utc=True)  # type: ignore[attr-defined]
file_handler.setLevel(LOG_LEVEL)
file_handler.setFormatter(logging.Formatter("%(message)s"))  # type: ignore[attr-defined]
logging.getLogger().addHandler(file_handler)  # type: ignore[attr-defined]
logging.getLogger().setLevel(LOG_LEVEL)  # type: ignore[attr-defined]

# Stdout logging for containers/cloud
if os.getenv("LOG_STDOUT", "0") == "1":
    stream_handler = logging.StreamHandler()  # type: ignore[attr-defined]
    stream_handler.setLevel(LOG_LEVEL)
    stream_handler.setFormatter(logging.Formatter("%(message)s"))  # type: ignore[attr-defined]
    logging.getLogger().addHandler(stream_handler)  # type: ignore[attr-defined]



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Gerenciador principal de conexões AMQP usando aio-pika."""

import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Any, cast

from rabbitmq_mcp_connection.connection.aio_pika_compat import (
    AbstractChannel,
    RobustConnection,
    aio_pika,
)
from rabbitmq_mcp_connection.connection.exceptions import (
    AuthenticationError,
    ConnectionError,
    ConnectionTimeoutError,
    VHostNotFoundError,
)
from rabbitmq_mcp_connection.connection.monitor import ConnectionMonitor
from rabbitmq_mcp_connection.logging.config import get_logger
from rabbitmq_mcp_connection.schemas.connection import (
    ConnectionConfig,
    ConnectionState,
    ConnectionStatusResult,
)
from rabbitmq_mcp_connection.schemas.retry import RetryPolicy

logger = get_logger(__name__)


@dataclass(slots=True)
class ConnectionMetadata:
    """Metadados sobre a conexão atual."""

    connected_since: datetime | None = None
    server_properties: dict[str, Any] | None = None


class ConnectionManager:
    """Gerencia conexão AMQP e reconexões automáticas."""

    def __init__(self, config: ConnectionConfig) -> None:
        self.config = config
        self.connection: RobustConnection | None = None
        self.channel: AbstractChannel | None = None
        self.state = ConnectionState.DISCONNECTED
        self.retry_policy = RetryPolicy()
        self._lock = asyncio.Lock()
        self._metadata = ConnectionMetadata()
        self._reconnect_task: asyncio.Task[None] | None = None
        self.monitor: ConnectionMonitor | None = None
        self._reconnect_callback_registered = False

    async def connect(self) -> None:
        """Estabelece uma conexão AMQP seguindo FR-001..FR-012."""

        async with self._lock:
            if self.state == ConnectionState.CONNECTED:
                return

            self.state = ConnectionState.CONNECTING
            logger.info(
                "connection.connecting",
                host=self.config.host,
                port=self.config.port,
                vhost=self.config.vhost,
            )

            try:
                self.connection = cast(
                    RobustConnection,
                    await asyncio.wait_for(
                        aio_pika.connect_robust(
                            host=self.config.host,
                            port=self.config.port,
                            login=self.config.user,
                            password=self.config.password,
                            virtualhost=self.config.vhost,
                            heartbeat=self.config.heartbeat,
                            timeout=self.config.timeout,
                        ),
                        timeout=self.config.timeout,
                    ),
                )
                self._reconnect_callback_registered = False
            except TimeoutError as exc:  # pragma: no cover - requires real broker
                self.state = ConnectionState.FAILED
                raise ConnectionTimeoutError("Tempo limite excedido ao conectar") from exc
            except aio_pika.exceptions.ProbableAuthenticationError as exc:
                self.state = ConnectionState.FAILED
                raise AuthenticationError("Credenciais inválidas para RabbitMQ") from exc
            except Exception as exc:  # pragma: no cover - erros externos
                self.state = ConnectionState.FAILED
                message = str(exc)
                if "not_allowed" in message.lower() or "vhost" in message.lower():
                    raise VHostNotFoundError(f"VHost não encontrado: {self.config.vhost}") from exc
                raise ConnectionError("Falha ao estabelecer conexão") from exc

            assert self.connection is not None
            self.channel = await self.connection.channel()  # canal default
            self.state = ConnectionState.CONNECTED
            self.retry_policy.reset()
            self._metadata.connected_since = datetime.now(UTC)
            self._metadata.server_properties = getattr(self.connection, "server_properties", None)

            if self.monitor is not None:
                await self.monitor.stop()
            if self.connection is not None:
                self.monitor = ConnectionMonitor(
                    self.connection,
                    on_connection_lost=self._handle_connection_lost,
                )
                await self.monitor.start()
                self._register_reconnect_callback()

            logger.info(
                "connection.established",
                host=self.config.host,
                port=self.config.port,
                vhost=self.config.vhost,
                connection_url=self.config.get_connection_url(),
            )

    async def disconnect(self) -> None:
        """Desconecta graciosamente fechando recursos conforme FR-007."""

        async with self._lock:
            if self.state == ConnectionState.DISCONNECTED:
                return

            logger.info("connection.disconnecting")

            if self._reconnect_task and not self._reconnect_task.done():
                self._reconnect_task.cancel()
                self._reconnect_task = None

            if self.monitor is not None:
                await self.monitor.stop()
                self.monitor = None

            self._unregister_reconnect_callback()

            if self.channel and not self.channel.is_closed:
                await self.channel.close()
            if self.connection and not self.connection.is_closed:
                await self.connection.close()

            self.channel = None
            self.connection = None
            self.state = ConnectionState.DISCONNECTED
            self._metadata = ConnectionMetadata()
            self._reconnect_callback_registered = False

            logger.info("connection.disconnected")

    async def get_status(self) -> ConnectionStatusResult:
        """Retorna informações sobre a conexão atual (FR-009)."""

        return ConnectionStatusResult(
            state=self.state,
            connection_url=self.config.get_connection_url(),
            connected_since=self._metadata.connected_since,
            retry_attempts=self.retry_policy.attempts,
            next_retry_in_seconds=(
                self.retry_policy.current_delay
                if self.state == ConnectionState.RECONNECTING
                else None
            ),
        )

    async def ensure_connected(self) -> None:
        """Garante que a conexão está ativa, reconectando se necessário."""

        if self.state == ConnectionState.CONNECTED:
            return
        await self.connect()

    async def reconnect(self) -> None:
        """Inicia processo de reconexão com backoff exponencial (FR-011)."""

        async with self._lock:
            if self.state == ConnectionState.RECONNECTING:
                return
            self.state = ConnectionState.RECONNECTING
            self.retry_policy.reset()

        logger.warning("connection.reconnecting", host=self.config.host)

        while True:
            try:
                await self.connect()
                logger.info("connection.reconnected")
                return
            except ConnectionError as exc:
                delay = self.retry_policy.next_delay()
                logger.error(
                    "connection.retry_failed",
                    error=str(exc),
                    attempt=self.retry_policy.attempts,
                    next_retry=delay,
                )
                await asyncio.sleep(delay)

    def schedule_reconnect(self) -> None:
        """Agenda reconexão em segundo plano."""

        if self._reconnect_task is None or self._reconnect_task.done():
            self._reconnect_task = asyncio.create_task(self.reconnect())

    async def _handle_connection_lost(self) -> None:
        logger.warning("connection.lost", host=self.config.host)
        monitor_to_stop: ConnectionMonitor | None = None
        async with self._lock:
            if self.state == ConnectionState.RECONNECTING:
                return
            monitor_to_stop = self.monitor
            self.state = ConnectionState.RECONNECTING
            self.channel = None
            self.monitor = None
            self._metadata.connected_since = None

        if monitor_to_stop is not None:
            await monitor_to_stop.stop()
        self._register_reconnect_callback()

    @asynccontextmanager
    async def channel_context(self) -> AsyncIterator[AbstractChannel]:
        """Context manager para uso seguro do canal."""

        await self.ensure_connected()
        assert self.channel is not None
        yield self.channel

    def __repr__(self) -> str:  # pragma: no cover - repr simples
        return f"ConnectionManager(state={self.state.value}, config={self.config!r})"

    def _register_reconnect_callback(self) -> None:
        if self.connection is None or self._reconnect_callback_registered:
            return
        callbacks = getattr(self.connection, "reconnect_callbacks", None)
        if callbacks is None or not hasattr(callbacks, "add"):
            logger.debug("connection.reconnect_callback_missing")
            return
        callbacks.add(self._on_connection_reconnected)
        self._reconnect_callback_registered = True

    def _unregister_reconnect_callback(self) -> None:
        if self.connection is None or not self._reconnect_callback_registered:
            return
        callbacks = getattr(self.connection, "reconnect_callbacks", None)
        if callbacks is not None and hasattr(callbacks, "discard"):
            callbacks.discard(self._on_connection_reconnected)
        self._reconnect_callback_registered = False

    async def _on_connection_reconnected(self, *_: Any, **__: Any) -> None:
        logger.info("connection.reconnected")
        new_monitor: ConnectionMonitor | None = None
        async with self._lock:
            if self.connection is None:
                return

            try:
                self.channel = await self.connection.channel()
            except Exception as exc:  # pragma: no cover - depende do broker real
                logger.error("connection.reconnect_channel_failed", error=str(exc))
                self.channel = None
                return

            self.state = ConnectionState.CONNECTED
            self.retry_policy.reset()
            self._metadata.connected_since = datetime.now(UTC)
            self._metadata.server_properties = getattr(self.connection, "server_properties", None)

            new_monitor = ConnectionMonitor(
                self.connection,
                on_connection_lost=self._handle_connection_lost,
            )
            self.monitor = new_monitor

        if new_monitor is not None:
            await new_monitor.start()



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

from datetime import UTC, datetime
from enum import Enum
from typing import Any

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)

__all__ = ["LogLevel", "LogCategory", "OperationResult", "LogEntry"]

MAX_MESSAGE_LENGTH = 100_000
TRUNCATION_SUFFIX = "...[truncated]"
SEMVER_PATTERN = r"^\d+\.\d+\.\d+$"
TOOL_NAME_PATTERN = r"^[a-z0-9_\-]+$"


class LogLevel(str, Enum):
    ERROR = "ERROR"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"


class LogCategory(str, Enum):
    CONNECTION = "CONNECTION"
    OPERATION = "OPERATION"
    ERROR = "ERROR"
    SECURITY = "SECURITY"
    PERFORMANCE = "PERFORMANCE"


class OperationResult(str, Enum):
    SUCCESS = "success"
    ERROR = "error"
    FAILURE = "error"
    TIMEOUT = "timeout"


class LogEntry(BaseModel):
    """Structured log entry validated against the documented schema."""

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    schema_version: str = Field(
        default="1.0.0",
        pattern=SEMVER_PATTERN,
        description="Semantic version for the log entry schema",
    )
    timestamp: str = Field(
        ...,
        description="ISO 8601 timestamp in UTC with trailing Z",
    )
    level: LogLevel
    category: LogCategory
    correlation_id: str | None = Field(None, min_length=1)
    message: str = Field(..., min_length=1)

    tool_name: str | None = Field(None, pattern=TOOL_NAME_PATTERN)
    operation_id: str | None = None
    duration_ms: float | None = Field(None, ge=0)
    operation_result: OperationResult | None = Field(  # accepts legacy "result"
        default=None,
        validation_alias=AliasChoices("operation_result", "result"),
    )
    error_type: str | None = None
    error_message: str | None = None
    stack_trace: str | None = None
    context: dict[str, Any] | None = None

    @property
    def result(self) -> OperationResult | None:
        """Expose legacy result alias for backwards compatibility."""

        return self.operation_result

    @field_validator("timestamp", mode="before")
    @classmethod
    def coerce_timestamp(cls, value: Any) -> str:
        if isinstance(value, datetime):
            aware = value if value.tzinfo else value.replace(tzinfo=UTC)
            utc_value = aware.astimezone(UTC)
            serialized = utc_value.isoformat().replace("+00:00", "Z")
            return serialized
        if isinstance(value, str):
            return value
        raise TypeError("timestamp must be str or datetime")

    @field_validator("timestamp")
    @classmethod
    def validate_timestamp(cls, value: str) -> str:
        if not value.endswith("Z"):
            raise ValueError("timestamp must be in UTC and end with 'Z'")
        try:
            datetime.fromisoformat(value.replace("Z", "+00:00"))
        except ValueError as exc:
            raise ValueError("timestamp must be ISO 8601 compliant") from exc
        return value

    @field_validator("schema_version")
    @classmethod
    def validate_schema_version(cls, value: str) -> str:
        major_str, *_ = value.split(".")
        try:
            major = int(major_str)
        except ValueError as exc:
            raise ValueError("schema_version major component must be numeric") from exc
        if major != 1:
            raise ValueError("unsupported schema_version major component for parser v1.0.0")
        return value

    @field_validator("message")
    @classmethod
    def truncate_message(cls, value: str) -> str:
        if len(value) <= MAX_MESSAGE_LENGTH:
            return value
        return value[:MAX_MESSAGE_LENGTH] + TRUNCATION_SUFFIX

    @field_validator("error_type")
    @classmethod
    def validate_error_type(cls, value: str | None, info: ValidationInfo) -> str | None:
        if value is None:
            return None
        level = info.data.get("level")
        if level is not None and level is not LogLevel.ERROR:
            raise ValueError("error_type requires level ERROR")
        return value

    @field_validator("error_message")
    @classmethod
    def validate_error_message(cls, value: str | None, info: ValidationInfo) -> str | None:
        if value is None:
            return None
        level = info.data.get("level")
        if level is not None and level is not LogLevel.ERROR:
            raise ValueError("error_message requires level ERROR")
        return value



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Pool assíncrono de conexões RabbitMQ."""

import asyncio
from datetime import UTC, datetime

from rabbitmq_mcp_connection.connection.exceptions import PoolTimeoutError
from rabbitmq_mcp_connection.connection.manager import ConnectionManager
from rabbitmq_mcp_connection.logging.config import get_logger
from rabbitmq_mcp_connection.schemas.connection import ConnectionConfig, PoolStats

LOGGER = get_logger(__name__)


class PooledConnection:
    """Wrapper de ``ConnectionManager`` rastreável para uso em pool."""

    def __init__(self, manager: ConnectionManager) -> None:
        self.manager = manager
        self.connection = manager.connection
        self.channel = manager.channel
        self.in_use = False
        self.created_at = datetime.now(UTC)
        self.last_used = self.created_at

    async def check_health(self) -> bool:
        connection = self.manager.connection
        channel = self.manager.channel
        if connection is None or getattr(connection, "is_closed", False):
            return False
        if channel is None or getattr(channel, "is_closed", False):
            return False
        return True

    async def reset(self) -> None:
        await self.manager.disconnect()
        await self.manager.connect()
        self.connection = self.manager.connection
        self.channel = self.manager.channel
        self.last_used = datetime.now(UTC)

    def mark_in_use(self) -> None:
        self.in_use = True
        self.last_used = datetime.now(UTC)

    def mark_available(self) -> None:
        self.in_use = False
        self.last_used = datetime.now(UTC)

    async def close(self) -> None:
        await self.manager.disconnect()


class ConnectionPool:
    """Pool assíncrono de conexões reutilizáveis."""

    def __init__(
        self,
        config: ConnectionConfig,
        *,
        max_size: int = 5,
        timeout: int = 10,
    ) -> None:
        if max_size < 1:
            raise ValueError("max_size must be >= 1")
        if timeout < 1:
            raise ValueError("timeout must be >= 1 second")

        self.config = config
        self.max_size = max_size
        self.timeout = timeout
        self._available: asyncio.Queue[PooledConnection] = asyncio.Queue(maxsize=max_size)
        self._all_connections: list[PooledConnection] = []
        self._lock = asyncio.Lock()
        self._waiters = 0

    async def initialize(self) -> None:
        async with self._lock:
            missing = self.max_size - len(self._all_connections)
            for _ in range(missing):
                pooled = await self._create_connection()
                self._all_connections.append(pooled)
                await self._available.put(pooled)
        LOGGER.info("pool.initialized", size=self.max_size)

    async def acquire(self) -> PooledConnection:
        self._waiters += 1
        try:
            pooled = await asyncio.wait_for(self._available.get(), timeout=self.timeout)
        except TimeoutError as exc:
            raise PoolTimeoutError(f"No connection available after {self.timeout}s") from exc
        finally:
            self._waiters = max(self._waiters - 1, 0)

        pooled.mark_in_use()
        LOGGER.debug("pool.connection_acquired", in_use=self.in_use_count)
        return pooled

    async def release(self, pooled: PooledConnection) -> None:
        if not isinstance(pooled, PooledConnection):  # pragma: no cover - defesa
            raise TypeError("release expects PooledConnection instance")

        if await pooled.check_health():
            pooled.mark_available()
            await self._available.put(pooled)
            LOGGER.debug("pool.connection_released", available=self.available_count)
            return

        await self._replace_connection(pooled)

    async def close(self) -> None:
        for pooled in self._all_connections:
            try:
                await pooled.close()
            except Exception:  # pragma: no cover - fechamento best-effort
                LOGGER.warning("pool.connection_close_failed")
        self._all_connections.clear()
        while not self._available.empty():
            self._available.get_nowait()
        LOGGER.info("pool.closed")

    async def get_stats(self) -> PoolStats:
        return PoolStats(
            max_size=self.max_size,
            total_connections=len(self._all_connections),
            in_use=self.in_use_count,
            available=self.available_count,
            waiting_for_connection=self._waiters,
            acquire_timeout_seconds=self.timeout,
        )

    @property
    def in_use_count(self) -> int:
        return sum(1 for conn in self._all_connections if conn.in_use)

    @property
    def available_count(self) -> int:
        return self._available.qsize()

    async def _create_connection(self) -> PooledConnection:
        manager = ConnectionManager(self.config)
        await manager.connect()
        return PooledConnection(manager)

    async def _replace_connection(self, pooled: PooledConnection) -> None:
        async with self._lock:
            if pooled in self._all_connections:
                self._all_connections.remove(pooled)
            try:
                await pooled.close()
            except Exception:  # pragma: no cover - fechamento best-effort
                LOGGER.warning("pool.connection_discard_failed")

            new_connection = await self._create_connection()
            self._all_connections.append(new_connection)
            await self._available.put(new_connection)
            LOGGER.info("pool.connection_replaced")


__all__ = ["ConnectionPool", "PooledConnection"]



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Manual sanity check for connection reconnection handling.

Stops the local RabbitMQ docker container mid-connection and verifies
that the connection manager transitions through RECONNECTING back to
CONNECTED once the container is started again.
"""

from __future__ import annotations

import asyncio
from asyncio.subprocess import PIPE

from rabbitmq_mcp_connection.connection.manager import ConnectionManager
from rabbitmq_mcp_connection.schemas.connection import ConnectionConfig, ConnectionState


async def run_cmd(*args: str) -> int:
    """Run a command and echo its stdout/stderr to the console."""

    proc = await asyncio.create_subprocess_exec(*args, stdout=PIPE, stderr=PIPE)
    stdout, stderr = await proc.communicate()
    if stdout:
        print(stdout.decode().strip())
    if stderr:
        print(stderr.decode().strip())
    return proc.returncode or 0


async def wait_for_state(
    manager: ConnectionManager, desired: ConnectionState, timeout: float
) -> None:
    """Poll the manager state until it matches *desired* or timeout expires."""

    loop = asyncio.get_event_loop()
    deadline = loop.time() + timeout
    while loop.time() < deadline:
        if manager.state == desired:
            return
        await asyncio.sleep(0.5)
    raise RuntimeError(
        f"Timed out waiting for state {desired.value}, current: {manager.state.value}"
    )


async def main() -> None:
    manager = ConnectionManager(ConnectionConfig())
    container_started = False

    await manager.connect()
    method = (
        "add_close_callback"
        if manager.monitor and manager.monitor._callback_collection is None
        else "close_callbacks"
    )
    print(f"monitor method: {method}")

    try:
        await asyncio.sleep(1.0)
        print("Stopping RabbitMQ container...")
        stop_rc = await run_cmd("docker", "stop", "rabbitmq")
        print(f"docker stop rc: {stop_rc}")

        try:
            await wait_for_state(manager, ConnectionState.RECONNECTING, timeout=10.0)
            print(f"manager state after stop: {manager.state.value}")
        except RuntimeError as exc:
            print(f"WARNING: {exc}")

        await asyncio.sleep(5.0)
        print("Starting RabbitMQ container...")
        start_rc = await run_cmd("docker", "start", "rabbitmq")
        print(f"docker start rc: {start_rc}")
        container_started = start_rc == 0

        try:
            await wait_for_state(manager, ConnectionState.CONNECTED, timeout=90.0)
            print(f"manager state after restart: {manager.state.value}")
        except RuntimeError as exc:
            print(f"ERROR: {exc}")
    finally:
        if not container_started:
            print("Ensuring container is running...")
            await run_cmd("docker", "start", "rabbitmq")
        await manager.disconnect()
        print(f"manager final state: {manager.state.value}")


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import pytest
from pytest import MonkeyPatch

from rabbitmq_mcp_connection.tools import search_ids
from rabbitmq_mcp_connection.tools.search_ids import (
    KeywordFallbackBackend,
    handle_search_ids,
)


@pytest.fixture(autouse=True)
def _force_keyword_backend(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(search_ids, "_BACKEND", KeywordFallbackBackend())


def test_handle_search_ids_requires_query() -> None:
    result = handle_search_ids("")
    assert result.success is False
    assert result.error is not None
    assert result.error.code == "INVALID_QUERY"


def test_handle_search_ids_returns_results() -> None:
    result = handle_search_ids("connection")
    assert result.success is True
    assert result.result is not None
    items = result.result["items"]
    assert len(items) > 0
    assert all("operation_id" in item for item in items)
    assert result.metadata == {"backend": "KeywordFallbackBackend"}



"""
This file is auto-generated from OpenAPI contracts.
LGPL v3.0 applies. See LICENSE in project root.
"""

from typing import Any

from pydantic import BaseModel, Field


class BindingCreateRequest(BaseModel):
    routing_key: str | None = Field(None)
    arguments: dict[str, Any] | None = Field(default=None)



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import io
import json
from contextlib import redirect_stderr

import pytest

from src.logging.handlers.console import ConsoleLogHandler


@pytest.fixture
def handler_and_buffer():
    buffer = io.StringIO()
    handler = ConsoleLogHandler(stream=buffer)
    return handler, buffer


def _read_output(buffer: io.StringIO):
    buffer.seek(0)
    return [json.loads(line) for line in buffer.read().splitlines() if line]


def test_console_handler_writes_to_stderr(handler_and_buffer):
    handler, buffer = handler_and_buffer

    handler.write_batch([{"event": "test"}])

    output = buffer.getvalue().strip()
    assert output


def test_console_handler_writes_json_format(handler_and_buffer):
    handler, buffer = handler_and_buffer

    handler.write_batch([{"event": "first"}, {"event": "second", "value": 2}])

    records = _read_output(buffer)
    assert records == [
        {"event": "first"},
        {"event": "second", "value": 2},
    ]


def test_console_handler_never_raises_exception():
    handler = ConsoleLogHandler()

    buffer = io.StringIO()
    with redirect_stderr(buffer):
        handler.write_batch([{"event": "test"}])

    with redirect_stderr(io.StringIO()):
        handler.write_batch([{"event": "test"}])



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""
MCP Tool: Semantic Search for Operation IDs
Implements search_ids(query: str, page: int, pageSize: int) -> SearchResult
"""

from pydantic import BaseModel

from tools.vector_db.search import semantic_search


class OperationMatch(BaseModel):
    operation_id: str
    score: float
    summary: str


class SearchResult(BaseModel):
    results: list[OperationMatch]
    page: int
    pageSize: int
    totalItems: int
    totalPages: int
    hasNextPage: bool
    hasPreviousPage: bool


def search_ids(query: str, page: int = 1, pageSize: int = 50) -> SearchResult:
    # Validate pagination
    if page < 1:
        raise ValueError("page must be >= 1")
    if not (1 <= pageSize <= 200):
        raise ValueError("pageSize must be between 1 and 200")
    # Run semantic search
    matches = semantic_search(query, limit=1000)  # Get all possible matches
    totalItems = len(matches)
    totalPages = max(1, (totalItems + pageSize - 1) // pageSize)
    start = (page - 1) * pageSize
    end = start + pageSize
    paged_matches = matches[start:end]
    hasNextPage = page < totalPages
    hasPreviousPage = page > 1
    # Convert to OperationMatch if needed
    results = [
        OperationMatch(**m) if not isinstance(m, OperationMatch) else m for m in paged_matches
    ]
    return SearchResult(
        results=results,
        page=page,
        pageSize=pageSize,
        totalItems=totalItems,
        totalPages=totalPages,
        hasNextPage=hasNextPage,
        hasPreviousPage=hasPreviousPage,
    )



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import time
from collections.abc import Callable, Iterable
from pathlib import Path

import pytest

import src.logging.logger as logger_module
from src.logging.handlers import file as file_module
from src.logging.logger import (
    StructuredLogger,
    get_logger,
    reset_structlog_configuration,
    shutdown_all_loggers,
)
from src.models.log_config import LogConfig, LogLevel
from src.models.log_entry import LogCategory


@pytest.fixture(autouse=True)
def reset_logging_state() -> Iterable[None]:
    shutdown_all_loggers()
    reset_structlog_configuration()
    yield
    shutdown_all_loggers()
    reset_structlog_configuration()


@pytest.fixture
def instrumentation(monkeypatch: pytest.MonkeyPatch) -> tuple[list[list[dict]], list[list[dict]]]:
    file_batches: list[list[dict]] = []
    console_batches: list[list[dict]] = []

    def fake_file_writer(self, batch: Iterable[dict]) -> None:  # type: ignore[override]
        file_batches.append(list(batch))

    def fake_console_writer(self, batch: Iterable[dict]) -> None:  # type: ignore[override]
        console_batches.append(list(batch))

    monkeypatch.setattr(file_module.FileLogHandler, "write_batch", fake_file_writer, raising=False)
    monkeypatch.setattr(
        logger_module.ConsoleLogHandler, "write_batch", fake_console_writer, raising=False
    )

    return file_batches, console_batches


@pytest.fixture
def logger_factory(tmp_path: Path) -> Callable[[str], StructuredLogger]:
    def _create(name: str) -> StructuredLogger:
        log_path = tmp_path / f"{name}.log"
        config = LogConfig(  # type: ignore[call-arg]
            output_file=str(log_path),
            log_level=LogLevel.DEBUG,
            async_flush_interval=0.2,
            batch_size=25,
            fallback_to_console=True,
        )
        logger = get_logger(name, config=config)
        return logger

    return _create


def _log_many(logger: StructuredLogger, *, total: int, category: LogCategory) -> None:
    for index in range(total):
        logger.info("performance test", category=category, entry=index)


def test_log_overhead_under_5ms(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    logger = logger_factory("overhead")
    file_batches, console_batches = instrumentation

    iterations = 200
    start = time.perf_counter()
    _log_many(logger, total=iterations, category=LogCategory.PERFORMANCE)
    elapsed = time.perf_counter() - start

    avg_ms = (elapsed / iterations) * 1000
    logger.flush()
    logger.shutdown()

    assert avg_ms < 5.0
    assert console_batches == []
    assert sum(len(batch) for batch in file_batches) == iterations


def test_async_logging_throughput(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    logger = logger_factory("throughput")
    file_batches, _ = instrumentation

    iterations = 1_000
    start = time.perf_counter()
    _log_many(logger, total=iterations, category=LogCategory.OPERATION)
    total_time = time.perf_counter() - start
    logger.flush()
    logger.shutdown()

    assert total_time < 1.0
    assert sum(len(batch) for batch in file_batches) == iterations


def test_batching_reduces_syscalls(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    logger = logger_factory("batching")
    file_batches, console_batches = instrumentation

    iterations = 125
    _log_many(logger, total=iterations, category=LogCategory.SECURITY)
    logger.flush()
    logger.shutdown()

    expected_max_calls = (iterations // 25) + 1
    assert len(file_batches) <= expected_max_calls
    assert console_batches == []


def test_performance_with_large_context_data(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    logger = logger_factory("large-context")
    file_batches, console_batches = instrumentation

    context_payload = {f"key_{i}": i for i in range(200)}

    iterations = 100
    start = time.perf_counter()
    for index in range(iterations):
        logger.info(
            "performance with context",
            category=LogCategory.OPERATION,
            context=context_payload,
            sequence=index,
        )
    elapsed = time.perf_counter() - start
    logger.flush()
    logger.shutdown()

    avg_ms = (elapsed / iterations) * 1000
    assert avg_ms < 8.0
    assert console_batches == []
    assert sum(len(batch) for batch in file_batches) == iterations


# T032: Integration tests for performance logging with timing helpers


def test_performance_logs_include_duration(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    """Test that logs generated with timing helpers include duration_ms field."""
    from src.logging.logger import log_duration

    logger = logger_factory("duration-test")
    file_batches, _ = instrumentation

    with log_duration(logger, "test_operation", user_id=123):
        time.sleep(0.1)

    logger.flush()
    logger.shutdown()

    assert len(file_batches) > 0
    all_logs = [log for batch in file_batches for log in batch]
    assert len(all_logs) == 1

    log_entry = all_logs[0]
    assert "duration_ms" in log_entry
    assert log_entry["duration_ms"] >= 100  # At least 100ms due to sleep
    assert log_entry["category"] == LogCategory.PERFORMANCE.value
    assert log_entry["operation_result"] == "success"
    assert log_entry["user_id"] == 123


def test_decorator_captures_operation_timing(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    """Test that @log_timing decorator captures operation timing correctly."""
    from src.logging.logger import log_timing

    file_batches, _ = instrumentation

    @log_timing(logger_name="decorator-test")
    def slow_operation() -> str:
        time.sleep(0.05)
        return "completed"

    # Create logger to ensure config is set
    logger_factory("decorator-test")

    result = slow_operation()

    # Get logger and flush
    logger = get_logger("decorator-test")
    logger.flush()
    logger.shutdown()

    assert result == "completed"

    all_logs = [log for batch in file_batches for log in batch]
    assert len(all_logs) == 1

    log_entry = all_logs[0]
    assert "duration_ms" in log_entry
    assert log_entry["duration_ms"] >= 50  # At least 50ms due to sleep
    assert log_entry["category"] == LogCategory.PERFORMANCE.value
    assert log_entry["operation_result"] == "success"
    assert "slow_operation" in log_entry["event"]


def test_context_manager_captures_timing(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    """Test that context manager captures timing for operations."""
    from src.logging.logger import log_duration

    logger = logger_factory("context-test")
    file_batches, _ = instrumentation

    with log_duration(logger, "database_query", level="debug", query_id=456):
        time.sleep(0.02)

    logger.flush()
    logger.shutdown()

    all_logs = [log for batch in file_batches for log in batch]
    assert len(all_logs) == 1

    log_entry = all_logs[0]
    assert "duration_ms" in log_entry
    assert log_entry["duration_ms"] >= 20
    assert log_entry["level"] == "DEBUG"
    assert log_entry["category"] == LogCategory.PERFORMANCE.value
    assert log_entry["query_id"] == 456


@pytest.mark.asyncio
async def test_async_operations_dont_block_logging(
    logger_factory: Callable[[str], StructuredLogger],
    instrumentation: tuple[list[list[dict]], list[list[dict]]],
) -> None:
    """Test that async operations with timing don't block the logging system."""
    import asyncio

    from src.logging.logger import log_timing

    file_batches, _ = instrumentation

    @log_timing(logger_name="async-test", level="info")
    async def async_operation() -> str:
        await asyncio.sleep(0.03)
        return "async_completed"

    # Create logger to ensure config is set
    logger_factory("async-test")

    # Run multiple async operations concurrently
    start = time.perf_counter()
    results = await asyncio.gather(
        async_operation(),
        async_operation(),
        async_operation(),
    )
    elapsed = time.perf_counter() - start

    # Get logger and flush
    logger = get_logger("async-test")
    logger.flush()
    logger.shutdown()

    # All operations should complete
    assert results == ["async_completed", "async_completed", "async_completed"]

    # Should take ~30ms (concurrent), not 90ms (sequential)
    assert elapsed < 0.06

    # Should have logged all 3 operations
    all_logs = [log for batch in file_batches for log in batch]
    assert len(all_logs) == 3

    for log_entry in all_logs:
        assert "duration_ms" in log_entry
        assert log_entry["category"] == LogCategory.PERFORMANCE.value
        assert log_entry["operation_result"] == "success"



#!/usr/bin/env python3
"""
Script to add LGPL v3.0 license headers to all Python source files.
Handles files with shebangs, encoding declarations, and existing headers.
"""

import argparse
import re
from pathlib import Path

LGPL_HEADER = '''"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""'''

# Patterns to detect existing license headers
LICENSE_PATTERNS = [
    r"LGPL v3\.0",
    r"GNU Lesser General Public License",
    r"GNU LESSER GENERAL PUBLIC LICENSE",
    r"Copyright \(C\)",
    r"This file is part of",
]

# Files to skip
SKIP_PATTERNS = [
    "*/__pycache__/*",
    "*/.*",  # Hidden files/directories
    "*.pyc",
    "*.pyo",
    "*/venv/*",
    "*/.venv/*",
    "*/build/*",
    "*/dist/*",
    "*/.pytest_cache/*",
    "*/.mypy_cache/*",
    "*/.ruff_cache/*",
    "*/htmlcov/*",
    "*/data/vectors/*",
]


def should_skip_file(file_path: Path) -> bool:
    """Check if file should be skipped based on patterns."""
    path_str = str(file_path)
    for pattern in SKIP_PATTERNS:
        if Path(path_str).match(pattern):
            return True
    return False


def has_license_header(content: str) -> bool:
    """Check if file already has a license header."""
    # Check only the first 30 lines for license patterns
    # Must appear in a docstring (triple-quoted) or comment, not in code
    lines = content.split("\n")[:30]

    # Look for our standard license header format
    # It should be in a docstring at the top of the file
    in_docstring = False
    docstring_content = []

    for line in lines:
        stripped = line.strip()

        # Skip shebang and encoding
        if stripped.startswith("#!") or "coding:" in stripped or "coding=" in stripped:
            continue

        # Start of docstring
        if not in_docstring and (stripped.startswith('"""') or stripped.startswith("'''")):
            in_docstring = True
            docstring_content.append(line)
            # Check if it's a single-line docstring
            if stripped.count('"""') >= 2 or stripped.count("'''") >= 2:
                in_docstring = False
            continue

        # Inside docstring
        if in_docstring:
            docstring_content.append(line)
            if '"""' in line or "'''" in line:
                in_docstring = False
                # Now check if this docstring contains license info
                docstring_text = "\n".join(docstring_content)
                for pattern in LICENSE_PATTERNS:
                    if re.search(pattern, docstring_text, re.IGNORECASE):
                        return True
                docstring_content = []

    return False


def extract_special_lines(content: str) -> tuple[list[str], str]:
    """
    Extract shebang and encoding declaration from the beginning of the file.
    Returns (special_lines, remaining_content).
    """
    lines = content.split("\n")
    special_lines = []
    idx = 0

    # Check for shebang (must be first line)
    if lines and lines[0].startswith("#!"):
        special_lines.append(lines[0])
        idx = 1

    # Check for encoding declaration (must be in first or second line)
    if idx < len(lines) and re.match(r"#.*coding[:=]\s*([-\w.]+)", lines[idx]):
        special_lines.append(lines[idx])
        idx += 1

    remaining_content = "\n".join(lines[idx:])
    return special_lines, remaining_content


def add_header_to_file(file_path: Path, dry_run: bool = False) -> bool:
    """
    Add license header to a Python file.
    Returns True if file was modified, False otherwise.
    """
    try:
        with open(file_path, encoding="utf-8") as f:
            content = f.read()
    except UnicodeDecodeError:
        print(f"⚠️  Skipping {file_path} (encoding issue)")
        return False

    # Skip if already has license
    if has_license_header(content):
        print(f"✓ {file_path} (already has license)")
        return False

    # Extract special lines (shebang, encoding)
    special_lines, remaining_content = extract_special_lines(content)

    # Remove leading empty lines from remaining content
    remaining_content = remaining_content.lstrip("\n")

    # Build new content
    new_lines = []

    # Add special lines first
    if special_lines:
        new_lines.extend(special_lines)
        new_lines.append("")  # Blank line after special lines

    # Add license header
    new_lines.append(LGPL_HEADER)

    # Add remaining content
    if remaining_content:
        new_lines.append("")  # Blank line after header
        new_lines.append(remaining_content)

    new_content = "\n".join(new_lines)

    if dry_run:
        print(f"🔍 Would add header to: {file_path}")
        return True

    # Write back to file
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(new_content)

    print(f"✅ Added header to: {file_path}")
    return True


def find_python_files(root_dir: Path) -> list[Path]:
    """Find all Python files in the directory tree."""
    python_files = []

    for pattern in ["**/*.py"]:
        for file_path in root_dir.glob(pattern):
            if file_path.is_file() and not should_skip_file(file_path):
                python_files.append(file_path)

    return sorted(python_files)


def main():
    parser = argparse.ArgumentParser(
        description="Add LGPL v3.0 license headers to Python source files"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would be done without making changes",
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help="Check if all files have license headers (exit 1 if any missing)",
    )
    parser.add_argument(
        "paths",
        nargs="*",
        type=Path,
        default=[Path("src"), Path("scripts"), Path("tests")],
        help="Paths to process (default: src scripts tests)",
    )

    args = parser.parse_args()

    # Collect all Python files
    all_files = []
    for path in args.paths:
        if not path.exists():
            print(f"⚠️  Path does not exist: {path}")
            continue

        if path.is_file():
            if path.suffix == ".py" and not should_skip_file(path):
                all_files.append(path)
        else:
            all_files.extend(find_python_files(path))

    if not all_files:
        print("No Python files found to process.")
        return 0

    print(f"Found {len(all_files)} Python files to check.\n")

    # Process files
    modified_count = 0
    missing_headers = []

    for file_path in all_files:
        was_modified = add_header_to_file(file_path, dry_run=args.dry_run or args.check)
        if was_modified:
            modified_count += 1
            if args.check:
                missing_headers.append(file_path)

    # Summary
    print(f"\n{'='*60}")
    if args.check:
        if missing_headers:
            print(f"❌ {len(missing_headers)} files missing license headers:")
            for path in missing_headers:
                print(f"   - {path}")
            return 1
        else:
            print("✅ All files have license headers!")
            return 0
    elif args.dry_run:
        print(f"Would add headers to {modified_count} files.")
        print("Run without --dry-run to apply changes.")
    else:
        print(f"✅ Added headers to {modified_count} files.")
        print(f"✓ {len(all_files) - modified_count} files already had headers.")

    return 0


if __name__ == "__main__":
    exit(main())



"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import json
from pathlib import Path

import pytest

from mcp_server.utils.logging import (
    SENSITIVE_FIELD_PATTERNS,
    _redact_string,
    configure_logging,
    reset_logging,
)


@pytest.fixture(autouse=True)
def _reset_logging_state():
    reset_logging()
    yield
    reset_logging()


def _read_records(handler):
    path = handler.current_log_path
    if path is None or not Path(path).exists():
        return []
    raw = Path(path).read_text(encoding="utf-8")
    return [json.loads(line) for line in raw.splitlines() if line.strip()]


def test_structured_logger_outputs_json_and_redacts_sensitive_fields(tmp_path):
    ctx = configure_logging(output_dir=tmp_path, app_name="rabbitmq-mcp-test")
    logger = ctx.logger

    payload = {pattern: f"{pattern}-secret" for pattern in SENSITIVE_FIELD_PATTERNS}
    message = " ".join(f"{pattern}=topsecret" for pattern in SENSITIVE_FIELD_PATTERNS)

    logger.info(message, **payload)
    ctx.flush()

    general_handler = next(handler for handler in ctx.handlers if handler.name == "general")
    records = _read_records(general_handler)
    assert records, "esperava pelo menos um registro de log"

    record = records[-1]
    assert record.get("message")
    assert "timestamp" in record
    assert record.get("level") in {"info", "INFO"}

    for pattern in SENSITIVE_FIELD_PATTERNS:
        assert record.get(pattern) == "***", f"{pattern} não foi redatado"
        assert f"{pattern}=topsecret" not in record["message"].lower()
        assert f"{pattern}=***" in _redact_string(record["message"]).lower()


def test_logger_respects_environment_log_level(tmp_path, monkeypatch):
    monkeypatch.setenv("LOG_LEVEL", "ERROR")
    ctx = configure_logging(output_dir=tmp_path, app_name="rabbitmq-mcp-test")
    logger = ctx.logger

    logger.info("mensagem ignorada", password="secret")
    ctx.flush()

    general_handler = next(handler for handler in ctx.handlers if handler.name == "general")
    records = _read_records(general_handler)
    assert not records, "logs INFO não deveriam ser registrados quando LOG_LEVEL=ERROR"


def test_error_logs_include_exception_context(tmp_path):
    ctx = configure_logging(output_dir=tmp_path, app_name="rabbitmq-mcp-test")
    logger = ctx.logger

    try:
        raise ValueError("falha proposital")
    except ValueError:
        logger.exception("falha na operação", request_id="req-123")

    ctx.flush()

    general_handler = next(handler for handler in ctx.handlers if handler.name == "general")
    records = _read_records(general_handler)
    assert records

    record = records[-1]
    assert record["message"] == "falha na operação"
    assert record["request_id"] == "req-123"
    assert "exception" in record
    assert "ValueError" in record["exception"]


def test_rotation_policy_metadata(tmp_path):
    ctx = configure_logging(output_dir=tmp_path, app_name="rabbitmq-mcp-test")

    policies = ctx.policies
    assert policies["general"]["retention_days"] == 30
    assert policies["general"]["max_bytes"] == 100 * 1024 * 1024
    assert policies["general"]["rotation"] == "daily"

    assert policies["errors"]["retention_days"] == 90
    assert policies["audit"]["retention_days"] == 365

    general_handler = next(handler for handler in ctx.handlers if handler.name == "general")
    assert getattr(general_handler, "max_bytes", None) == 100 * 1024 * 1024
    assert getattr(general_handler, "retention_days", None) == 30



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from typing import Any
from urllib.parse import quote

import requests
from requests import Session
from requests.exceptions import HTTPError

from utils.errors import (
    AuthorizationError,
    ConflictError,
    ConnectionError,
    NotFoundError,
    RabbitMQError,
    ValidationError,
)


class RabbitMQExecutor:
    """HTTP client responsible for interacting with the RabbitMQ Management API."""

    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        vhost: str = "/",
        use_tls: bool = False,
        timeout: float = 5.0,
    ) -> None:
        scheme = "https" if use_tls else "http"
        self.base_url = f"{scheme}://{host}:{port}/api"
        self.session = Session()
        self.session.auth = (user, password)
        self.vhost = vhost
        self.timeout = timeout
        self.logger = logging.getLogger("RabbitMQExecutor")  # type: ignore[attr-defined]

    def _encode_vhost(self, vhost: str) -> str:
        return quote(vhost, safe="")

    def _request(self, method: str, endpoint: str, **kwargs: Any) -> Any:
        url = f"{self.base_url}{endpoint}"
        try:
            response = self.session.request(method, url, timeout=self.timeout, **kwargs)
        except requests.RequestException as exc:  # pragma: no cover - network failure path
            self.logger.error("HTTP request failed", exc_info=exc)
            raise ConnectionError(
                code="NETWORK_ERROR",
                message="Unable to reach RabbitMQ Management API",
                action="Verify network connectivity and API availability",
                context={"url": url, "method": method},
            ) from exc

        status_code = response.status_code
        if status_code in (401, 403):
            self.logger.warning("Unauthorized request to %s", endpoint)
            raise AuthorizationError(
                code="UNAUTHORIZED",
                message="RabbitMQ credentials were rejected",
                action="Verify username, password and permissions",
                context={"endpoint": endpoint, "status": status_code},
            )

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error_context = {
                "endpoint": endpoint,
                "status": status_code,
                "method": method,
            }
            if status_code == 404:
                raise NotFoundError(
                    code="NOT_FOUND",
                    message="Requested resource was not found",
                    action="Ensure the target resource exists before retrying",
                    context=error_context,
                ) from exc
            if status_code == 409:
                raise ConflictError(
                    code="CONFLICT",
                    message="RabbitMQ reported a resource conflict",
                    action="Resolve existing resource state before retrying",
                    context=error_context,
                ) from exc
            if 400 <= status_code < 500:
                raise ValidationError(
                    code=f"HTTP_{status_code}",
                    message=response.text or "RabbitMQ rejected the request",
                    action="Adjust request parameters and retry",
                    context=error_context,
                ) from exc
            raise ConnectionError(
                code="REMOTE_ERROR",
                message="RabbitMQ Management API returned an error",
                action="Retry later or review RabbitMQ server logs",
                context=error_context,
            ) from exc

        if status_code == 204:
            return None

        if not response.content:
            return None

        try:
            return response.json()
        except ValueError as exc:  # pragma: no cover - unexpected content type
            raise RabbitMQError(
                code="INVALID_RESPONSE",
                message="RabbitMQ responded with non-JSON payload",
                action="Inspect response payload for debugging",
                context={"endpoint": endpoint},
            ) from exc

    def get_queues(self, vhost: str | None = None) -> Any:
        if vhost is None:
            endpoint = "/queues"
        else:
            endpoint = f"/queues/{self._encode_vhost(vhost)}"
        return self._request("GET", endpoint)

    def get_exchanges(self, vhost: str | None = None) -> Any:
        if vhost is None:
            endpoint = "/exchanges"
        else:
            endpoint = f"/exchanges/{self._encode_vhost(vhost)}"
        return self._request("GET", endpoint)

    def get_bindings(self, vhost: str | None = None) -> Any:
        if vhost is None:
            endpoint = "/bindings"
        else:
            endpoint = f"/bindings/{self._encode_vhost(vhost)}"
        return self._request("GET", endpoint)

    def delete_queue(self, vhost: str, queue: str, force: bool = False) -> Any:
        params = {"if-empty": "false"} if force else {}
        endpoint = f"/queues/{self._encode_vhost(vhost)}/{quote(queue, safe='')}"
        return self._request("DELETE", endpoint, params=params)

    def get_queue(self, vhost: str, name: str) -> Any:
        endpoint = f"/queues/{self._encode_vhost(vhost)}/{quote(name, safe='')}"
        return self._request("GET", endpoint)

    def create_queue(self, vhost: str, name: str, payload: dict[str, Any]) -> Any:
        endpoint = f"/queues/{self._encode_vhost(vhost)}/{quote(name, safe='')}"
        return self._request("PUT", endpoint, json=payload)

    def get_exchange(self, vhost: str, name: str) -> Any:
        endpoint = f"/exchanges/{self._encode_vhost(vhost)}/{quote(name, safe='')}"
        return self._request("GET", endpoint)

    def create_exchange(self, vhost: str, name: str, payload: dict[str, Any]) -> Any:
        endpoint = f"/exchanges/{self._encode_vhost(vhost)}/{quote(name, safe='')}"
        return self._request("PUT", endpoint, json=payload)

    def delete_exchange(self, vhost: str, name: str) -> Any:
        endpoint = f"/exchanges/{self._encode_vhost(vhost)}/{quote(name, safe='')}"
        return self._request("DELETE", endpoint)

    def create_binding(
        self,
        vhost: str,
        exchange: str,
        queue: str,
        payload: dict[str, Any],
    ) -> Any:
        endpoint = (
            f"/bindings/{self._encode_vhost(vhost)}/e/{quote(exchange, safe='')}"
            f"/q/{quote(queue, safe='')}"
        )
        return self._request("POST", endpoint, json=payload)

    def list_binding_relations(self, vhost: str, exchange: str, queue: str) -> Any:
        endpoint = (
            f"/bindings/{self._encode_vhost(vhost)}/e/{quote(exchange, safe='')}"
            f"/q/{quote(queue, safe='')}"
        )
        return self._request("GET", endpoint)

    def get_bindings_for_exchange(self, vhost: str, exchange: str) -> Any:
        endpoint = (
            f"/exchanges/{self._encode_vhost(vhost)}/" f"{quote(exchange, safe='')}/bindings/source"
        )
        return self._request("GET", endpoint)

    def delete_binding(
        self,
        vhost: str,
        exchange: str,
        queue: str,
        properties_key: str,
    ) -> Any:
        endpoint = (
            f"/bindings/{self._encode_vhost(vhost)}/e/{quote(exchange, safe='')}"
            f"/q/{quote(queue, safe='')}/{quote(properties_key, safe='')}"
        )
        return self._request("DELETE", endpoint)



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

"""Shared helpers for CLI command options and execution."""

from collections.abc import Callable, Iterable, Sequence
from typing import Any

import click

from cli.formatters.json import format_json
from cli.formatters.table import format_table
from tools import call_id as call_id_tool

ConnectionDict = dict[str, Any]
PaginationDict = dict[str, Any]


def connection_options(func: Callable[..., Any]) -> Callable[..., Any]:
    options = [
        click.option(
            "--password",
            envvar="RABBITMQ_PASSWORD",
            required=True,
            hide_input=True,
            help="RabbitMQ password",
        ),
        click.option(
            "--user",
            envvar="RABBITMQ_USER",
            required=True,
            help="RabbitMQ username",
        ),
        click.option(
            "--port",
            envvar="RABBITMQ_PORT",
            default=15672,
            show_default=True,
            type=int,
            help="RabbitMQ management port",
        ),
        click.option(
            "--host",
            envvar="RABBITMQ_HOST",
            required=True,
            help="RabbitMQ host",
        ),
        click.option(
            "--use-tls/--no-use-tls",
            envvar="RABBITMQ_USE_TLS",
            default=False,
            show_default=True,
            help="Use HTTPS when contacting the management API",
        ),
        click.option(
            "--connection-vhost",
            envvar="RABBITMQ_CONNECTION_VHOST",
            default="/",
            show_default=True,
            help="Virtual host used for authentication context",
        ),
    ]

    for option in reversed(options):
        func = option(func)
    return func


def pagination_options(func: Callable[..., Any]) -> Callable[..., Any]:
    options = [
        click.option(
            "--page-size",
            default=50,
            show_default=True,
            type=int,
            help="Items per page (1-200)",
        ),
        click.option(
            "--page",
            default=1,
            show_default=True,
            type=int,
            help="Page number (1-based)",
        ),
    ]

    for option in reversed(options):
        func = option(func)
    return func


def output_options(func: Callable[..., Any]) -> Callable[..., Any]:
    options = [
        click.option(
            "--format",
            "output_format",
            type=click.Choice(["table", "json"], case_sensitive=False),
            default="table",
            show_default=True,
            help="Output format",
        ),
        click.option(
            "--verbose/--no-verbose",
            default=False,
            show_default=True,
            help="Include additional statistics in table output",
        ),
    ]

    for option in reversed(options):
        func = option(func)
    return func


def build_connection_payload(
    *,
    host: str,
    port: int,
    user: str,
    password: str,
    connection_vhost: str,
    use_tls: bool,
    vhost: str | None,
) -> ConnectionDict:
    payload: ConnectionDict = {
        "host": host,
        "port": port,
        "user": user,
        "password": password,
        "connection_vhost": connection_vhost or "/",
        "use_tls": use_tls,
    }
    if vhost is not None:
        payload["vhost"] = vhost
    return payload


def build_pagination_payload(*, page: int, page_size: int) -> PaginationDict:
    return {"page": page, "pageSize": page_size}


def execute_operation(
    endpoint: str,
    params: ConnectionDict,
    pagination: PaginationDict | None = None,
) -> dict[str, Any]:
    response = call_id_tool.call_id(endpoint, params=params, pagination=pagination)
    if hasattr(response, "model_dump"):
        return response.model_dump()
    if isinstance(response, dict):
        return response
    raise TypeError(f"Unexpected response type: {type(response)!r}")


def format_table_output(
    items: Iterable[dict[str, Any]],
    columns: Sequence[tuple[str, str]],
    footer: str,
) -> str:
    headers = [header for header, _ in columns]
    rows: list[dict[str, Any]] = []
    for item in items:
        row: dict[str, Any] = {}
        for header, key in columns:
            row[header] = _extract_value(item, key)
        rows.append(row)
    table = format_table(rows, headers)
    return f"{table}\n{footer}".strip()


def format_json_output(payload: dict[str, Any]) -> str:
    return format_json(payload)


def _extract_value(item: dict[str, Any], dotted_key: str) -> Any:
    value: Any = item
    for part in dotted_key.split("."):
        if isinstance(value, dict):
            value = value.get(part)
        else:
            value = getattr(value, part, None)
    return value


__all__ = [
    "connection_options",
    "pagination_options",
    "output_options",
    "build_connection_payload",
    "build_pagination_payload",
    "execute_operation",
    "format_table_output",
    "format_json_output",
]



from __future__ import annotations

"""
Copyright (C) 2025 Luciano Guerche

This file is part of rabbitmq-mcp-server.

rabbitmq-mcp-server is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rabbitmq-mcp-server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with rabbitmq-mcp-server. If not, see <https://www.gnu.org/licenses/>.
"""

import asyncio
import atexit
import functools
import logging
import signal
import sys
import threading
import time
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, TypeVar, cast

import orjson
import structlog
import structlog.stdlib
import structlog.types

from models.log_config import LogConfig, LogLevel
from models.log_entry import LogCategory
from utils.async_writer import AsyncLogWriter

from .handlers.console import ConsoleLogHandler
from .handlers.file import FileLogHandler
from .processors import add_correlation_id, redact_sensitive_data

__all__ = [
    "configure_structlog",
    "get_logger",
    "log_duration",
    "log_timing",
    "reset_structlog_configuration",
    "shutdown_all_loggers",
    "StructuredLogger",
]

F = TypeVar("F", bound=Callable[..., Any])


_STRUCTLOG_CONFIGURED = False
_LOGGER_REGISTRY: dict[str, _LoggerState] = {}
_REGISTRY_LOCK = threading.Lock()
_LOGGER_NAME_FIELD = "_logger_name"
_SIGNAL_HANDLERS_REGISTERED = False

_METHOD_TO_LEVEL = {
    "error": LogLevel.ERROR,
    "exception": LogLevel.ERROR,
    "warning": LogLevel.WARN,
    "warn": LogLevel.WARN,
    "info": LogLevel.INFO,
    "debug": LogLevel.DEBUG,
}

_LEVEL_TO_NUM = {
    LogLevel.ERROR: logging.ERROR,
    LogLevel.WARN: logging.WARNING,
    LogLevel.INFO: logging.INFO,
    LogLevel.DEBUG: logging.DEBUG,
}


def _json_serializer(obj: structlog.types.EventDict, **kwargs: Any) -> str:
    default = kwargs.pop("default", str)
    event = dict(obj)

    timestamp = event.get("timestamp")
    if isinstance(timestamp, str) and timestamp.endswith("+00:00"):
        event["timestamp"] = f"{timestamp[:-6]}Z"

    event.setdefault("schema_version", "1.0.0")

    dispatched = _dispatch_to_log_manager(None, "", event)
    return orjson.dumps(dispatched, default=default).decode("utf-8")


def _dispatch_to_log_manager(
    logger: Any, method_name: str, event_dict: dict[str, Any]
) -> dict[str, Any]:
    manager_name = event_dict.pop(_LOGGER_NAME_FIELD, None)
    if not manager_name:
        return event_dict

    state = _LOGGER_REGISTRY.get(manager_name)
    if not state:
        return event_dict

    try:
        state.emit(dict(event_dict))
    except Exception:
        # Fallback to console to avoid losing logs if file handling breaks unexpectedly.
        state.console_handler.write_batch([dict(event_dict)])
    return event_dict


def configure_structlog(*, logger_factory: Any | None = None) -> None:
    global _STRUCTLOG_CONFIGURED
    if _STRUCTLOG_CONFIGURED:
        return

    if logger_factory is None:
        logger_factory = structlog.PrintLoggerFactory()

    processor_chain = [
        cast(
            structlog.types.Processor,
            structlog.processors.TimeStamper(fmt="iso", utc=True, key="timestamp"),
        ),
        cast(structlog.types.Processor, add_correlation_id),
        cast(structlog.types.Processor, redact_sensitive_data),
        cast(structlog.types.Processor, structlog.processors.StackInfoRenderer()),
        cast(structlog.types.Processor, structlog.processors.format_exc_info),
        cast(
            structlog.types.Processor,
            structlog.processors.JSONRenderer(serializer=_json_serializer),
        ),
    ]

    structlog.configure(
        processors=processor_chain,
        wrapper_class=structlog.stdlib.BoundLogger,
        context_class=dict,
        logger_factory=logger_factory,
        cache_logger_on_first_use=True,
    )

    _STRUCTLOG_CONFIGURED = True


def get_logger(name: str = "rabbitmq-mcp", *, config: LogConfig | None = None) -> StructuredLogger:
    """
    Get or create a structured logger.

    Automatically registers signal handlers for graceful shutdown on first call.

    Args:
        name: Logger name (default: "rabbitmq-mcp")
        config: Optional logger configuration

    Returns:
        StructuredLogger instance with configured handlers
    """
    configure_structlog()
    _register_signal_handlers()

    existing_config: LogConfig | None = None
    with _REGISTRY_LOCK:
        current_state = _LOGGER_REGISTRY.get(name)
        if current_state is not None:
            existing_config = current_state.config

    if config is None:
        config = existing_config if existing_config is not None else LogConfig()  # type: ignore[call-arg]

    state = _get_or_create_state(name, config)
    bound = structlog.get_logger(name)
    return StructuredLogger(bound, state)


def reset_structlog_configuration() -> None:
    global _STRUCTLOG_CONFIGURED
    shutdown_all_loggers()
    structlog.reset_defaults()
    _STRUCTLOG_CONFIGURED = False


def shutdown_all_loggers() -> None:
    """
    Gracefully shutdown all active loggers, flushing buffered logs.

    This function ensures zero log loss by blocking until all buffered
    logs are written to their destinations. Called automatically on:
    - Normal program exit (via atexit)
    - SIGTERM signal (graceful shutdown request)
    - SIGINT signal (Ctrl+C)

    Thread-safe and idempotent.
    """
    _shutdown_all_loggers()


def _shutdown_all_loggers() -> None:
    with _REGISTRY_LOCK:
        states = list(_LOGGER_REGISTRY.values())
        _LOGGER_REGISTRY.clear()
    for state in states:
        state.shutdown()


def _register_signal_handlers() -> None:
    """
    Register signal handlers for graceful shutdown.

    Registers handlers for:
    - SIGTERM: Graceful shutdown (Docker, systemd, kill)
    - SIGINT: Ctrl+C in terminal
    - atexit: Normal program exit

    Only registers once (idempotent).
    """
    global _SIGNAL_HANDLERS_REGISTERED

    if _SIGNAL_HANDLERS_REGISTERED:
        return

    # Register atexit handler for normal program exit
    atexit.register(_shutdown_all_loggers)

    # Register signal handlers (Unix/Linux/macOS only)
    if sys.platform != "win32":

        def _signal_handler(signum: int, frame: Any) -> None:
            """Handle shutdown signals gracefully."""
            _shutdown_all_loggers()
            # Re-raise signal to allow default handler to run
            signal.signal(signum, signal.SIG_DFL)
            signal.raise_signal(signum)

        # SIGTERM: Graceful shutdown request
        signal.signal(signal.SIGTERM, _signal_handler)

        # SIGINT: Ctrl+C
        signal.signal(signal.SIGINT, _signal_handler)

    _SIGNAL_HANDLERS_REGISTERED = True


def _get_or_create_state(name: str, config: LogConfig) -> _LoggerState:
    with _REGISTRY_LOCK:
        existing = _LOGGER_REGISTRY.get(name)
        if existing is not None:
            existing.ensure_compatible(config)
            return existing

        state = _LoggerState(name, config)
        _LOGGER_REGISTRY[name] = state
        return state


def _resolve_output_path(config: LogConfig) -> Path:
    template = config.output_file
    if "{date}" in template:
        today = datetime.utcnow().strftime("%Y-%m-%d")
        template = template.format(date=today)
    return Path(template).expanduser().resolve()


class _AsyncWriterRunner:
    def __init__(self, writer: AsyncLogWriter) -> None:
        self._writer = writer
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._started = False
        self._start_lock = threading.Lock()

    def start(self) -> None:
        with self._start_lock:
            if self._started:
                return
            self._thread.start()
            asyncio.run_coroutine_threadsafe(self._writer.start(), self._loop).result()
            self._started = True

    def submit(self, entry: dict[str, Any]) -> None:
        if not self._started:
            raise RuntimeError("Async writer not started")
        asyncio.run_coroutine_threadsafe(self._writer.write_log(entry), self._loop).result()

    def flush(self) -> None:
        if not self._started:
            return
        asyncio.run_coroutine_threadsafe(self._writer.flush(), self._loop).result()

    def drain_pending(self) -> list[dict[str, Any]]:
        return self._writer.drain_pending()

    def stop(self, *, timeout: float = 30.0) -> None:
        """
        Stop the async writer with graceful shutdown.

        Args:
            timeout: Maximum time to wait for flush completion (seconds).
        """
        if not self._started:
            return
        try:
            asyncio.run_coroutine_threadsafe(self._writer.stop(timeout=timeout), self._loop).result(
                timeout=timeout + 1.0
            )
        except Exception:
            # Force stop if graceful shutdown fails
            pass
        finally:
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join(timeout=2.0)
            if self._thread.is_alive():
                # Thread didn't stop - log warning but continue
                pass
            self._loop.close()
            self._started = False

    def _run_loop(self) -> None:
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()


class _LoggerState:
    def __init__(self, name: str, config: LogConfig) -> None:
        self.name = name
        self.config = config
        self.console_handler = ConsoleLogHandler()

        fallback = self.console_handler.write_batch if config.fallback_to_console else None
        self.file_path = _resolve_output_path(config)
        self.file_handler = FileLogHandler(self.file_path, fallback=fallback, config=config)

        # Initialize RabbitMQ handler if enabled
        self.rabbitmq_handler: Any | None = None
        if config.rabbitmq.enabled:
            try:
                from .handlers.rabbitmq import RabbitMQLogHandler

                self.rabbitmq_handler = RabbitMQLogHandler(
                    host=config.rabbitmq.host,
                    port=config.rabbitmq.port,
                    vhost=config.rabbitmq.vhost,
                    username=config.rabbitmq.username,
                    password=config.rabbitmq.password,
                    exchange=config.rabbitmq.exchange,
                    exchange_type=config.rabbitmq.exchange_type,
                    durable=config.rabbitmq.durable,
                    fallback_to_console=config.fallback_to_console,
                )
            except Exception as exc:
                print(
                    f"WARNING: Failed to initialize RabbitMQ handler: {exc}. "
                    f"RabbitMQ logging disabled.",
                    file=sys.stderr,
                )

        async def _write_batch(entries: Iterable[dict[str, Any]]) -> None:
            batch = list(entries)
            if not batch:
                return

            # Write to file handler
            try:
                await asyncio.to_thread(self.file_handler.write_batch, batch)
            except Exception:
                await asyncio.to_thread(self.console_handler.write_batch, batch)

            # Write to RabbitMQ handler if enabled
            if self.rabbitmq_handler is not None:
                try:
                    await asyncio.to_thread(self.rabbitmq_handler.write_batch, batch)
                except Exception:
                    # Failure logged by RabbitMQ handler itself
                    pass

        self._writer = AsyncLogWriter(
            max_queue_size=config.async_queue_size,
            flush_interval=config.async_flush_interval,
            batch_size=config.batch_size,
            write_batch=_write_batch,
        )
        self._runner = _AsyncWriterRunner(self._writer)
        self._lock = threading.Lock()
        self._started = False

    def ensure_compatible(self, config: LogConfig) -> None:
        if config.model_dump() != self.config.model_dump():
            raise ValueError(f"Logger '{self.name}' already configured with different settings.")

    def ensure_started(self) -> None:
        with self._lock:
            if self._started:
                return
            self._runner.start()
            self._started = True

    def set_log_level(self, level: LogLevel) -> None:
        """
        Change the minimum log level at runtime (thread-safe).

        Args:
            level: New minimum log level
        """
        with self._lock:
            self.config.log_level = level

    def emit(self, event: dict[str, Any]) -> None:
        self.ensure_started()
        self._runner.submit(dict(event))

    def flush(self) -> None:
        if not self._started:
            return
        self._runner.flush()

    def shutdown(self) -> None:
        pending_entries: list[dict[str, Any]] = []
        with self._lock:
            if not self._started:
                pending_entries = self._runner.drain_pending()
            else:
                self._runner.stop()
                self._started = False
                pending_entries = self._runner.drain_pending()

        if pending_entries:
            try:
                self.file_handler.write_batch(pending_entries)
            except Exception:
                self.console_handler.write_batch(pending_entries)
            if self.rabbitmq_handler is not None:
                try:
                    self.rabbitmq_handler.write_batch(pending_entries)
                except Exception:
                    pass

        self.file_handler.flush()
        self.file_handler.close()
        if self.rabbitmq_handler is not None:
            try:
                self.rabbitmq_handler.close()
            except Exception:
                pass


class StructuredLogger:
    """Adapter that enforces structured logging semantics for this project."""

    __slots__ = ("_logger", "_state")

    def __init__(self, logger: Any, state: _LoggerState) -> None:
        self._logger = logger
        self._state = state

    def bind(self, **new_values: Any) -> StructuredLogger:
        return StructuredLogger(self._logger.bind(**new_values), self._state)

    def new(self, **new_values: Any) -> StructuredLogger:
        return StructuredLogger(self._logger.new(**new_values), self._state)

    def debug(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        return self._log("debug", message, category=category, **kwargs)

    def info(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        return self._log("info", message, category=category, **kwargs)

    def warning(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        return self._log("warning", message, category=category, **kwargs)

    def warn(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        return self.warning(message, category=category, **kwargs)

    def error(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        return self._log("error", message, category=category, **kwargs)

    def exception(self, message: str, *, category: LogCategory | str, **kwargs: Any) -> Any:
        kwargs.setdefault("exc_info", True)
        return self._log("exception", message, category=category, **kwargs)

    def flush(self) -> None:
        self._state.flush()

    def shutdown(self) -> None:
        self._state.shutdown()

    def set_log_level(self, level: LogLevel) -> None:
        """
        Change the minimum log level at runtime.

        Args:
            level: New minimum log level (DEBUG, INFO, WARN, ERROR)
        """
        self._state.set_log_level(level)

    def __getattr__(self, item: str) -> Any:
        return getattr(self._logger, item)

    def _log(
        self, method_name: str, message: str, *, category: LogCategory | str, **kwargs: Any
    ) -> Any:
        if category is None:
            raise ValueError("category is required for structured logging")

        log_level = _METHOD_TO_LEVEL.get(method_name)
        if log_level is None:
            raise ValueError(f"Unsupported log method '{method_name}'")

        if _LEVEL_TO_NUM[log_level] < _LEVEL_TO_NUM[self._state.config.log_level]:
            return None

        if "category" in kwargs:
            raise ValueError("'category' must be provided as named argument, not inside kwargs")

        category_value = category.value if isinstance(category, LogCategory) else str(category)

        payload = dict(kwargs)
        payload["category"] = category_value
        payload[_LOGGER_NAME_FIELD] = self._state.name

        return getattr(self._logger, method_name)(message, **payload)


@contextmanager
def log_duration(
    logger: StructuredLogger,
    operation: str,
    *,
    level: str = "info",
    category: LogCategory | str = LogCategory.PERFORMANCE,
    **kwargs: Any,
) -> Any:
    """
    Context manager for automatically logging operation duration.

    Usage:
        with log_duration(logger, "database_query", user_id=123):
            # ... operation ...
            pass

    Args:
        logger: StructuredLogger instance to use
        operation: Human-readable operation name
        level: Log level (debug, info, warning, error)
        category: Log category (defaults to Performance)
        **kwargs: Additional fields to include in the log entry
    """
    start_time = time.perf_counter()
    try:
        yield
        duration_ms = (time.perf_counter() - start_time) * 1000
        log_method = getattr(logger, level)
        log_method(
            f"Operation completed: {operation}",
            category=category,
            duration_ms=round(duration_ms, 2),
            operation_result="success",
            **kwargs,
        )
    except Exception as exc:
        duration_ms = (time.perf_counter() - start_time) * 1000
        logger.error(
            f"Operation failed: {operation}",
            category=category,
            duration_ms=round(duration_ms, 2),
            operation_result="error",
            error_type=type(exc).__name__,
            error_message=str(exc),
            **kwargs,
        )
        raise


def log_timing(
    operation: str | None = None,
    *,
    logger_name: str = "rabbitmq-mcp",
    level: str = "info",
    category: LogCategory | str = LogCategory.PERFORMANCE,
    **log_kwargs: Any,
) -> Callable[[F], F]:
    """
    Decorator for automatically logging function execution duration.

    Usage:
        @log_timing
        async def slow_operation():
            await asyncio.sleep(0.1)

        @log_timing(operation="custom_name", level="debug")
        def fast_operation():
            pass

    Args:
        operation: Operation name (defaults to function name)
        logger_name: Logger name to use
        level: Log level (debug, info, warning, error)
        category: Log category (defaults to Performance)
        **log_kwargs: Additional fields to include in log entries
    """

    def decorator(func: F) -> F:
        op_name = operation if operation is not None else func.__name__

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            logger = get_logger(logger_name)
            start_time = time.perf_counter()
            try:
                result = await func(*args, **kwargs)
                duration_ms = (time.perf_counter() - start_time) * 1000
                log_method = getattr(logger, level)
                log_method(
                    f"Operation completed: {op_name}",
                    category=category,
                    duration_ms=round(duration_ms, 2),
                    operation_result="success",
                    **log_kwargs,
                )
                return result
            except Exception as exc:
                duration_ms = (time.perf_counter() - start_time) * 1000
                logger.error(
                    f"Operation failed: {op_name}",
                    category=category,
                    duration_ms=round(duration_ms, 2),
                    operation_result="error",
                    error_type=type(exc).__name__,
                    error_message=str(exc),
                    **log_kwargs,
                )
                raise

        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            logger = get_logger(logger_name)
            start_time = time.perf_counter()
            try:
                result = func(*args, **kwargs)
                duration_ms = (time.perf_counter() - start_time) * 1000
                log_method = getattr(logger, level)
                log_method(
                    f"Operation completed: {op_name}",
                    category=category,
                    duration_ms=round(duration_ms, 2),
                    operation_result="success",
                    **log_kwargs,
                )
                return result
            except Exception as exc:
                duration_ms = (time.perf_counter() - start_time) * 1000
                logger.error(
                    f"Operation failed: {op_name}",
                    category=category,
                    duration_ms=round(duration_ms, 2),
                    operation_result="error",
                    error_type=type(exc).__name__,
                    error_message=str(exc),
                    **log_kwargs,
                )
                raise

        # Return async wrapper if function is coroutine, sync wrapper otherwise
        if asyncio.iscoroutinefunction(func):
            return async_wrapper  # type: ignore[return-value]
        else:
            return sync_wrapper  # type: ignore[return-value]

    return decorator


atexit.register(_shutdown_all_loggers)
